import contextlib
from locale import D_FMT
import sys
import contextlib

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "olorenchemengine"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

from rdkit import RDLogger

RDLogger.DisableLog("rdApp.*")

import logging

logging.getLogger().setLevel(logging.ERROR)

import os
from os import path

if not path.exists(path.join(path.expanduser("~"), f".oce/")):
    os.mkdir(path.join(path.expanduser("~"), f".oce/"))

if not path.exists(path.join(path.expanduser("~"), f".oce/cache/")):
    os.mkdir(path.join(path.expanduser("~"), f".oce/cache/"))

if not path.exists(path.join(path.expanduser("~"), f".oce/cache/vecrep/")):
    os.mkdir(path.join(path.expanduser("~"), f".oce/cache/vecrep/"))

import pandas as pd

import json


CONFIG_PATH = os.path.join(os.path.expanduser("~"), ".oce/CONFIG.json")
if os.path.exists(CONFIG_PATH):
    with open(CONFIG_PATH) as f:
        CONFIG_ = json.load(f)
else:
    CONFIG_ = {"MAP_LOCATION": "cpu", "USE_CUDA": False, "CDD_TOKEN": None, "VAULT_ID": None, "NUM_WORKERS": 4}
    with open(CONFIG_PATH, "w+") as f:
        json.dump(CONFIG_, f)

CONFIG = CONFIG_.copy()

def update_config():
    """Update the configuration file.

    This function is called when a new parameter is added to the configuration file.
    """

    global CONFIG
    CONFIG = CONFIG_.copy()

    with contextlib.suppress(ImportError):
        import torch
        CONFIG["DEVICE"] = torch.device(CONFIG["MAP_LOCATION"])
        CONFIG["USE_CUDA"] = "cuda" in CONFIG["MAP_LOCATION"]

    with open(CONFIG_PATH, "w+") as f:
        json.dump(CONFIG_, f)


def set_config_param(param, value):
    """Set a configuration parameter.

    Parameters:
        param: the parameter to set.
        value: the value to set the parameter to.
    """
    CONFIG_[param] = value
    update_config()


def remove_config_param(param):
    """Remove a configuration parameter.

    Parameters:
        param: the parameter to remove.
    """
    CONFIG_.pop(param)
    update_config()


update_config()

def ExampleDataFrame():
    return pd.read_csv("gs://oloren-public-data/sample-csvs/sample_data3.csv")

from .base_class import *
from .basics import *
from .ensemble import *
from .gnn import *
from .representations import *
from .external import *
from .splitters import *
from .dataset import *
from .interpret import *
from .manager import *
from .visualizations import *
from .uncertainty import *
from .benchmarks import *
from .hyperparameters import *

def ExampleDataset():
    if os.path.exists(path.join(path.expanduser("~"), f".oce/exampledataset.oce")):
        return load(path.join(path.expanduser("~"), f".oce/exampledataset.oce"))
    else:
        dataset =  BaseDataset(data = ExampleDataFrame().to_csv(), structure_col = "Smiles",
            property_col = "pChEMBL Value") + RandomSplit()
        save(dataset, path.join(path.expanduser("~"), f".oce/exampledataset.oce"))
        return dataset

def BACEDataset():
    df = pd.read_csv(download_public_file("MoleculeNet/load_bace_regression.csv"))
    df["split"] = df["split"].replace({"Train": "train", "Valid": "valid", "Test": "test"})
    return oce.BaseDataset(data = df.to_csv(), structure_col = "smiles", property_col = "pIC50")

def test_oce():
    """Convenience function to test all functions of the oce package."""

    df = ExampleDataFrame()

    model = BaseBoosting(
        [
            RandomForestModel(DescriptastorusDescriptor("rdkit2dnormalized"), n_estimators=1),
            BaseTorchGeometricModel(TLFromCheckpoint("default"), batch_size=8, epochs=1, preinitialized=True),
            RandomForestModel(OlorenCheckpoint("default"), n_estimators=1),
        ]
    )

    model.fit(df["Smiles"], df["pChEMBL Value"])
    _ = model.predict(["CC(=O)OC1=CC=CC=C1C(=O)O"])
    save(model, "model.oce")
    _ = load("model.oce")
    os.remove("model.oce")




#  COPYRIGHT Aspuru-Guzik group (2021)
#                                  Apache License
#                            Version 2.0, January 2004
#                         http://www.apache.org/licenses/

#    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION

#    1. Definitions.

#       "License" shall mean the terms and conditions for use, reproduction,
#       and distribution as defined by Sections 1 through 9 of this document.

#       "Licensor" shall mean the copyright owner or entity authorized by
#       the copyright owner that is granting the License.

#       "Legal Entity" shall mean the union of the acting entity and all
#       other entities that control, are controlled by, or are under common
#       control with that entity. For the purposes of this definition,
#       "control" means (i) the power, direct or indirect, to cause the
#       direction or management of such entity, whether by contract or
#       otherwise, or (ii) ownership of fifty percent (50%) or more of the
#       outstanding shares, or (iii) beneficial ownership of such entity.

#       "You" (or "Your") shall mean an individual or Legal Entity
#       exercising permissions granted by this License.

#       "Source" form shall mean the preferred form for making modifications,
#       including but not limited to software source code, documentation
#       source, and configuration files.

#       "Object" form shall mean any form resulting from mechanical
#       transformation or translation of a Source form, including but
#       not limited to compiled object code, generated documentation,
#       and conversions to other media types.

#       "Work" shall mean the work of authorship, whether in Source or
#       Object form, made available under the License, as indicated by a
#       copyright notice that is included in or attached to the work
#       (an example is provided in the Appendix below).

#       "Derivative Works" shall mean any work, whether in Source or Object
#       form, that is based on (or derived from) the Work and for which the
#       editorial revisions, annotations, elaborations, or other modifications
#       represent, as a whole, an original work of authorship. For the purposes
#       of this License, Derivative Works shall not include works that remain
#       separable from, or merely link (or bind by name) to the interfaces of,
#       the Work and Derivative Works thereof.

#       "Contribution" shall mean any work of authorship, including
#       the original version of the Work and any modifications or additions
#       to that Work or Derivative Works thereof, that is intentionally
#       submitted to Licensor for inclusion in the Work by the copyright owner
#       or by an individual or Legal Entity authorized to submit on behalf of
#       the copyright owner. For the purposes of this definition, "submitted"
#       means any form of electronic, verbal, or written communication sent
#       to the Licensor or its representatives, including but not limited to
#       communication on electronic mailing lists, source code control systems,
#       and issue tracking systems that are managed by, or on behalf of, the
#       Licensor for the purpose of discussing and improving the Work, but
#       excluding communication that is conspicuously marked or otherwise
#       designated in writing by the copyright owner as "Not a Contribution."

#       "Contributor" shall mean Licensor and any individual or Legal Entity
#       on behalf of whom a Contribution has been received by Licensor and
#       subsequently incorporated within the Work.

#    2. Grant of Copyright License. Subject to the terms and conditions of
#       this License, each Contributor hereby grants to You a perpetual,
#       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#       copyright license to reproduce, prepare Derivative Works of,
#       publicly display, publicly perform, sublicense, and distribute the
#       Work and such Derivative Works in Source or Object form.

#    3. Grant of Patent License. Subject to the terms and conditions of
#       this License, each Contributor hereby grants to You a perpetual,
#       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#       (except as stated in this section) patent license to make, have made,
#       use, offer to sell, sell, import, and otherwise transfer the Work,
#       where such license applies only to those patent claims licensable
#       by such Contributor that are necessarily infringed by their
#       Contribution(s) alone or by combination of their Contribution(s)
#       with the Work to which such Contribution(s) was submitted. If You
#       institute patent litigation against any entity (including a
#       cross-claim or counterclaim in a lawsuit) alleging that the Work
#       or a Contribution incorporated within the Work constitutes direct
#       or contributory patent infringement, then any patent licenses
#       granted to You under this License for that Work shall terminate
#       as of the date such litigation is filed.

#    4. Redistribution. You may reproduce and distribute copies of the
#       Work or Derivative Works thereof in any medium, with or without
#       modifications, and in Source or Object form, provided that You
#       meet the following conditions:

#       (a) You must give any other recipients of the Work or
#           Derivative Works a copy of this License; and

#       (b) You must cause any modified files to carry prominent notices
#           stating that You changed the files; and

#       (c) You must retain, in the Source form of any Derivative Works
#           that You distribute, all copyright, patent, trademark, and
#           attribution notices from the Source form of the Work,
#           excluding those notices that do not pertain to any part of
#           the Derivative Works; and

#       (d) If the Work includes a "NOTICE" text file as part of its
#           distribution, then any Derivative Works that You distribute must
#           include a readable copy of the attribution notices contained
#           within such NOTICE file, excluding those notices that do not
#           pertain to any part of the Derivative Works, in at least one
#           of the following places: within a NOTICE text file distributed
#           as part of the Derivative Works; within the Source form or
#           documentation, if provided along with the Derivative Works; or,
#           within a display generated by the Derivative Works, if and
#           wherever such third-party notices normally appear. The contents
#           of the NOTICE file are for informational purposes only and
#           do not modify the License. You may add Your own attribution
#           notices within Derivative Works that You distribute, alongside
#           or as an addendum to the NOTICE text from the Work, provided
#           that such additional attribution notices cannot be construed
#           as modifying the License.

#       You may add Your own copyright statement to Your modifications and
#       may provide additional or different license terms and conditions
#       for use, reproduction, or distribution of Your modifications, or
#       for any such Derivative Works as a whole, provided Your use,
#       reproduction, and distribution of the Work otherwise complies with
#       the conditions stated in this License.

#    5. Submission of Contributions. Unless You explicitly state otherwise,
#       any Contribution intentionally submitted for inclusion in the Work
#       by You to the Licensor shall be under the terms and conditions of
#       this License, without any additional terms or conditions.
#       Notwithstanding the above, nothing herein shall supersede or modify
#       the terms of any separate license agreement you may have executed
#       with Licensor regarding such Contributions.

#    6. Trademarks. This License does not grant permission to use the trade
#       names, trademarks, service marks, or product names of the Licensor,
#       except as required for reasonable and customary use in describing the
#       origin of the Work and reproducing the content of the NOTICE file.

#    7. Disclaimer of Warranty. Unless required by applicable law or
#       agreed to in writing, Licensor provides the Work (and each
#       Contributor provides its Contributions) on an "AS IS" BASIS,
#       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#       implied, including, without limitation, any warranties or conditions
#       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
#       PARTICULAR PURPOSE. You are solely responsible for determining the
#       appropriateness of using or redistributing the Work and assume any
#       risks associated with Your exercise of permissions under this License.

#    8. Limitation of Liability. In no event and under no legal theory,
#       whether in tort (including negligence), contract, or otherwise,
#       unless required by applicable law (such as deliberate and grossly
#       negligent acts) or agreed to in writing, shall any Contributor be
#       liable to You for damages, including any direct, indirect, special,
#       incidental, or consequential damages of any character arising as a
#       result of this License or out of the use or inability to use the
#       Work (including but not limited to damages for loss of goodwill,
#       work stoppage, computer failure or malfunction, or any and all
#       other commercial damages or losses), even if such Contributor
#       has been advised of the possibility of such damages.

#    9. Accepting Warranty or Additional Liability. While redistributing
#       the Work or Derivative Works thereof, You may choose to offer,
#       and charge a fee for, acceptance of support, warranty, indemnity,
#       or other liability obligations and/or rights consistent with this
#       License. However, in accepting such obligations, You may act only
#       on Your own behalf and on Your sole responsibility, not on behalf
#       of any other Contributor, and only if You agree to indemnify,
#       defend, and hold each Contributor harmless for any liability
#       incurred by, or claims asserted against, such Contributor by reason
#       of your accepting any such warranty or additional liability.

#    END OF TERMS AND CONDITIONS

#    APPENDIX: How to apply the Apache License to your work.

#       To apply the Apache License to your work, attach the following
#       boilerplate notice, with the fields enclosed by brackets "[]"
#       replaced with your own identifying information. (Don't include
#       the brackets!)  The text should be enclosed in the appropriate
#       comment syntax for the file format. We also recommend that a
#       file or class name and description of purpose be included on the
#       same "printed page" as the copyright notice for easier
#       identification within third-party archives.

#    Copyright [yyyy] [name of copyright owner]

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

#   CHANGES
#   1. Simplified imports
#   2. Added optional passed in index
#   3. Added condition to sanitize SMILES

import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.DataStructs.cDataStructs import TanimotoSimilarity
from rdkit.Chem import Mol
from rdkit.Chem.AtomPairs.Sheridan import GetBPFingerprint, GetBTFingerprint
from rdkit.Chem.Pharm2D import Generate, Gobbi_Pharm2D

from rdkit import RDLogger

import olorenchemengine as oce

RDLogger.DisableLog("rdApp.*")

def randomize_smiles(mol):
    '''Returns a random (dearomatized) SMILES given an rdkit mol object of a molecule.
    Parameters:
    mol (rdkit.Chem.rdchem.Mol) :  RdKit mol object (None if invalid smile string smi)
    
    Returns:
    mol (rdkit.Chem.rdchem.Mol) : RdKit mol object  (None if invalid smile string smi)
    '''
    if not mol:
        return None

    Chem.Kekulize(mol)
    return Chem.MolToSmiles(mol, canonical=False, doRandom=True, isomericSmiles=False,  kekuleSmiles=True)

def sanitize_smiles(smi):
    '''Return a canonical smile representation of smi
    
    Parameters:
    smi (string) : smile string to be canonicalized 
    
    Returns:
    mol (rdkit.Chem.rdchem.Mol) : RdKit mol object                          (None if invalid smile string smi)
    smi_canon (string)          : Canonicalized smile representation of smi (None if invalid smile string smi)
    conversion_successful (bool): True/False to indicate if conversion was  successful 
    '''
    try:
        mol = Chem.MolFromSmiles(smi, sanitize=True)
        smi_canon = Chem.MolToSmiles(mol, isomericSmiles=False, canonical=True)
        if Chem.MolFromSmiles(smi_canon) is None:
            return (None, None, False)
        return (mol, smi_canon, True)
    except:
        return (None, None, False)
    

def get_selfie_chars(selfie):
    '''Obtain a list of all selfie characters in string selfie
    
    Parameters: 
    selfie (string) : A selfie string - representing a molecule 
    
    Example: 
    >>> get_selfie_chars('[C][=C][C][=C][C][=C][Ring1][Branch1_1]')
    ['[C]', '[=C]', '[C]', '[=C]', '[C]', '[=C]', '[Ring1]', '[Branch1_1]']
    
    Returns:
    chars_selfie: list of selfie characters present in molecule selfie
    '''
    chars_selfie = [] # A list of all SELFIE sybols from string selfie
    while selfie != '':
        chars_selfie.append(selfie[selfie.find('['): selfie.find(']')+1])
        selfie = selfie[selfie.find(']')+1:]
    return chars_selfie

def mutate_selfie(selfie, max_molecules_len, write_fail_cases=False, index=None):
    '''Return a mutated selfie string (only one mutation on selfie is performed)
    
    Mutations are done until a valid molecule is obtained 
    Rules of mutation: With a 33.3% propbabily, either: 
        1. Add a random SELFIE character in the string
        2. Replace a random SELFIE character with another
        3. Delete a random character
    
    Parameters:
    selfie            (string)  : SELFIE string to be mutated 
    max_molecules_len (int)     : Mutations of SELFIE string are allowed up to this length
    write_fail_cases  (bool)    : If true, failed mutations are recorded in "selfie_failure_cases.txt"
    
    Returns:
    selfie_mutated    (string)  : Mutated SELFIE string
    smiles_canon      (string)  : canonical smile of mutated SELFIE string
    '''
    valid=False
    fail_counter = 0
    chars_selfie = get_selfie_chars(selfie)
    
    while not valid:
        fail_counter += 1
        
        oce.import_or_install("selfies")
        
        import selfies

        alphabet = list(selfies.get_semantic_robust_alphabet()) # 34 SELFIE characters 

        choice_ls = [1, 2, 3] # 1=Insert; 2=Replace; 3=Delete
        random_choice = np.random.choice(choice_ls, 1)[0]
        
        # Insert a character in a Random Location
        if random_choice == 1: 
            if index is None:
                random_index = np.random.randint(len(chars_selfie)+1)
            else:
                random_index = index
            random_character = np.random.choice(alphabet, size=1)[0]
            
            selfie_mutated_chars = chars_selfie[:random_index] + [random_character] + chars_selfie[random_index:]

        # Replace a random character 
        elif random_choice == 2:
            if index is None:                         
                random_index = np.random.randint(len(chars_selfie))
            else:
                random_index = index
            random_character = np.random.choice(alphabet, size=1)[0]
            if random_index == 0:
                selfie_mutated_chars = [random_character] + chars_selfie[random_index+1:]
            else:
                selfie_mutated_chars = chars_selfie[:random_index] + [random_character] + chars_selfie[random_index+1:]
                
        # Delete a random character
        elif random_choice == 3:
            if index is None: 
                random_index = np.random.randint(len(chars_selfie))
            else:
                random_index = index
            if random_index == 0:
                selfie_mutated_chars = chars_selfie[random_index+1:]
            else:
                selfie_mutated_chars = chars_selfie[:random_index] + chars_selfie[random_index+1:]
                
        else: 
            raise Exception('Invalid Operation trying to be performed')

        selfie_mutated = "".join(x for x in selfie_mutated_chars)
        sf = "".join(x for x in chars_selfie)
        
        try:
            smiles = decoder(selfie_mutated)
            mol, smiles_canon, done = sanitize_smiles(smiles)
            if len(selfie_mutated_chars) > max_molecules_len or smiles_canon=="":
                done = False
            if done:
                valid = True
            else:
                valid = False
        except:
            valid=False
            if fail_counter > 1 and write_fail_cases == True:
                f = open("selfie_failure_cases.txt", "a+")
                f.write('Tried to mutate SELFIE: '+str(sf)+' To Obtain: '+str(selfie_mutated) + '\n')
                f.close()
    
    return (selfie_mutated, smiles_canon)

def get_mutated_SELFIES(selfies_ls, num_mutations): 
    ''' Mutate all the SELFIES in 'selfies_ls' 'num_mutations' number of times. 
    
    Parameters:
    selfies_ls   (list)  : A list of SELFIES 
    num_mutations (int)  : number of mutations to perform on each SELFIES within 'selfies_ls'
    
    Returns:
    selfies_ls   (list)  : A list of mutated SELFIES
    
    '''
    for _ in range(num_mutations): 
        selfie_ls_mut_ls = []
        for str_ in selfies_ls: 
            
            str_chars = get_selfie_chars(str_)
            max_molecules_len = len(str_chars) + num_mutations
            
            selfie_mutated, _ = mutate_selfie(str_, max_molecules_len)
            selfie_ls_mut_ls.append(selfie_mutated)
        
        selfies_ls = selfie_ls_mut_ls.copy()
    return selfies_ls



"""
Wrapping the GIN network family based on:

`GINEPLUS GitHub repository <https://github.com/RBrossard/GINEPLUS>`_
`Graph convolutions that can finally model local structure <https://arxiv.org/abs/2011.15069>`_
"""
from olorenchemengine.base_class import log_arguments
from olorenchemengine.gnn import BaseLightningModule
from .operations import GINNetwork

class GINModel(BaseLightningModule):
    
    """
    GINModel class from `GINEPLUS GitHub repository <https://github.com/RBrossard/GINEPLUS>`.
    
    Parameters:
        task_type (str): Type of task to perform. Options are 'regression' or 'classification'. This
            task_type will automatically be set by BaseTorchGeometricModel and usually
            does not need to be set manually.
        hidden (int): Number of nodes per hidden layer.
        lr (float): Learning rate.
        layers (int): Number of hidden layers.
        dropout (float): Dropout rate.
        virtual_node (bool): Whether to use virtual node.
        conv_type (str): Type of convolution to use. Options are 'gcn', 'gin', 'gin+', 'naivegin+'.
        conv_radius (int): Radius of convolution.
        optim (str): Optimizer to use. Options are 'adam'.
    """
    @log_arguments
    def __init__(self, task_type = "classification", hidden: str = 100, lr: float = 0.001,
                 layers=3, dropout=0.5, virtual_node=False,
                 conv_radius=3, conv_type='gin+',optim="adam",**kwargs):
        assert conv_type in ['gcn', 'gin', 'gin+', 'naivegin+']
        super().__init__(optim=optim)

        # Network
        out_dim = 1
        self.network = GINNetwork(hidden=hidden,
                                         out_dim=out_dim,
                                         layers=layers,
                                         dropout=dropout,
                                         virtual_node=virtual_node,
                                         k=conv_radius,
                                         conv_type=conv_type)


import olorenchemengine as oce

from olorenchemengine.base_class import log_arguments, BaseModel
from olorenchemengine.representations import AtomFeaturizer, BondFeaturizer, TorchGeometricGraph, SMILESRepresentation, BaseVecRepresentation
from olorenchemengine.internal import download_public_file

from rdkit import Chem
from rdkit.Chem.rdchem import BondType as BT

try:
    from torch_geometric.data import DataLoader
except ImportError:
    oce.mock_imports(globals(), "DataLoader")

import numpy as np
from tqdm import tqdm

ATOM_LIST = list(range(1,119))
CHIRALITY_LIST = [
    Chem.rdchem.ChiralType.CHI_UNSPECIFIED,
    Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
    Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW,
    Chem.rdchem.ChiralType.CHI_OTHER
]
BOND_LIST = [BT.SINGLE, BT.DOUBLE, BT.TRIPLE, BT.AROMATIC]
BONDDIR_LIST = [
    Chem.rdchem.BondDir.NONE,
    Chem.rdchem.BondDir.ENDUPRIGHT,
    Chem.rdchem.BondDir.ENDDOWNRIGHT
]

class MolCLR_AF(AtomFeaturizer):
    @property
    def length(self):
        return 2

    def convert(self, atom: Chem.Atom):
        atom_feature = [ATOM_LIST.index(atom.GetAtomicNum()), CHIRALITY_LIST.index(atom.GetChiralTag())]
        x = np.array(atom_feature)
        return x

class MolCLR_BF(BondFeaturizer):
    @property
    def length(self):
        return 2

    def convert(self, bond: Chem.Bond):
        edge_feature = [
                BOND_LIST.index(bond.GetBondType()),
                BONDDIR_LIST.index(bond.GetBondDir())
            ]
        x = np.array(edge_feature)
        return x

class MolCLR_PYG(TorchGeometricGraph):
    @log_arguments
    def __init__(self):
        super().__init__(MolCLR_AF(), MolCLR_BF(), log = False)

class MolCLR(BaseModel):

    model_config = {"num_layer": 5,      # number of graph conv layers
            "emb_dim": 300,                  # embedding dimension in graph conv layers
            "feat_dim": 512,                 # output feature dimention
            "drop_ratio": 0.3,               # dropout ratio
            "pool": "mean"}

    @log_arguments
    def __init__(self, model_type = "ginet", epochs = 100, batch_size = 32,
                 init_lr = 0.0005, init_base_lr = 0.0001, weight_decay = 1e-6,
                 **kwargs):
        self.model_type = model_type

        self.config = {
            "epochs": epochs,
            "batch_size": batch_size,
            "init_lr": init_lr,
            "init_base_lr": init_base_lr,
            "weight_decay": weight_decay
        }

        self.representation = MolCLR_PYG()

        super().__init__(**kwargs)

    def preprocess(self, X, y, **kwargs):
        if y is None:
            y = [None]*len(X)
        return self.representation.convert(X, ys=y)

    def _fit(self, X, y = None, **kwargs):

        import torch
        from torch import nn
        import torch.nn.functional as F
        from torch.utils.tensorboard import SummaryWriter
        from torch.optim.lr_scheduler import CosineAnnealingLR
        from sklearn.metrics import roc_auc_score, mean_squared_error, mean_absolute_error

        if self.setting == 'regression':
            self.criterion = nn.MSELoss()
        elif self.setting == 'classification':
            self.criterion = nn.BCEWithLogitsLoss(reduction = "none")

        if self.model_type == "ginet":
            from .model import GINet
            self.model = GINet(self.setting, **self.model_config)
            save_path = download_public_file("MolCLR/pretrained_gin.pth")
        elif self.model_type == "gcn":
            from .model import GCN
            self.model = GCN(self.setting, **self.model_config)
            save_path = download_public_file("MolCLR/pretrained_gcn.pth")

        self.model.load_my_state_dict(torch.load(save_path, map_location = oce.CONFIG["MAP_LOCATION"]))
        self.model.to(oce.CONFIG["DEVICE"])

        layer_list = []
        for name, param in self.model.named_parameters():
            if 'pred_lin' in name:
                layer_list.append(name)

        params = list(map(lambda x: x[1],list(filter(lambda kv: kv[0] in layer_list, self.model.named_parameters()))))
        base_params = list(map(lambda x: x[1],list(filter(lambda kv: kv[0] not in layer_list, self.model.named_parameters()))))

        optimizer = torch.optim.Adam(
            [{'params': base_params, 'lr': self.config['init_base_lr']}, {'params': params}],
            self.config['init_lr'], weight_decay=self.config['weight_decay']
        )

        train_loader = DataLoader(X, batch_size=self.config["batch_size"], shuffle=True, num_workers=oce.CONFIG["NUM_WORKERS"])
        for epoch_counter in range(self.config['epochs']):
            print(f"Epoch {epoch_counter+1}/{self.config['epochs']}")
            for bn, batch in enumerate(train_loader):

                batch = batch.to(oce.CONFIG["DEVICE"])
                _, pred = self.model(batch)
                y = batch.y.view(pred.shape).to(torch.float64)

                #Loss matrix
                loss_mat = self.criterion(pred.double(), y)

                optimizer.zero_grad()

                loss = torch.mean(loss_mat)
                loss.backward()

                optimizer.step()

    def _step(self, model, data, n_iter):
        # get the prediction
        __, pred = model(data)  # [N,C]
        loss = self.criterion(pred.flatten(), data.y)

        return loss

    def _predict(self, X, **kwargs):
        import torch

        loader = DataLoader(X, batch_size=self.config["batch_size"], shuffle=False, num_workers=oce.CONFIG["NUM_WORKERS"])
        self.model.eval()
        y_pred = []

        for step, batch in enumerate(tqdm(loader, desc="Iteration")):
            batch = batch.to(oce.CONFIG["DEVICE"])

            with torch.no_grad():
                _, pred = self.model(batch)
            y_pred.append(pred.cpu().detach().numpy())

        return np.vstack(y_pred).flatten()

class MolCLRVecRep(BaseVecRepresentation):

    model_config = {"num_layer": 5,      # number of graph conv layers
            "emb_dim": 300,                  # embedding dimension in graph conv layers
            "feat_dim": 512,                 # output feature dimention
            "drop_ratio": 0.3,               # dropout ratio
            "pool": "mean"}

    @log_arguments
    def __init__(self, model_type = "ginet", epochs = 100, batch_size = 32,
                 init_lr = 0.0005, init_base_lr = 0.0001, weight_decay = 1e-6,
                 **kwargs):
        self.model_type = model_type

        self.config = {
            "epochs": epochs,
            "batch_size": batch_size,
            "init_lr": init_lr,
            "init_base_lr": init_base_lr,
            "weight_decay": weight_decay
        }

        import torch
        from torch import nn
        import torch.nn.functional as F
        from torch.utils.tensorboard import SummaryWriter
        from torch.optim.lr_scheduler import CosineAnnealingLR
        from sklearn.metrics import roc_auc_score, mean_squared_error, mean_absolute_error

        if self.model_type == "ginet":
            from .model import GINet
            self.model = GINet("regression", **self.model_config)
            save_path = download_public_file("MolCLR/pretrained_gin.pth")
        elif self.model_type == "gcn":
            from .model import GCN
            self.model = GCN("regression", **self.model_config)
            save_path = download_public_file("MolCLR/pretrained_gcn.pth")

        self.model.load_my_state_dict(torch.load(save_path, map_location = oce.CONFIG["MAP_LOCATION"]))
        self.model.to(oce.CONFIG["DEVICE"])

    def _convert(self, X, **kwargs):
        return self.convert(X)

    def convert(self, X, **kwargs):
        X = MolCLR_PYG().convert(X, ys=None)
        import torch

        loader = DataLoader(X, batch_size=self.config["batch_size"], shuffle=True, num_workers=oce.CONFIG["NUM_WORKERS"])
        self.model.eval()
        y_pred = []

        for step, batch in enumerate(tqdm(loader, desc="Iteration")):
            batch = batch.to(oce.CONFIG["DEVICE"])

            with torch.no_grad():
                rep, _ = self.model(batch)
            y_pred.append(rep.cpu().detach().numpy())

        return np.concatenate(y_pred)


import os

import numpy as np

from typing import List, Union, Any
from rdkit import Chem

import olorenchemengine as oce

from olorenchemengine.base_class import log_arguments
from olorenchemengine.representations import BaseVecRepresentation
from olorenchemengine.internal import download_public_file
from .operations import MolSentence, sentences2vec, mol2alt_sentence, DfVec

class Mol2Vec(BaseVecRepresentation):

    @log_arguments
    def __init__(self):
        print("Requires gensim installation https://github.com/RaRe-Technologies/gensim")

        model_path = download_public_file("mol2vec/model_300dim.pkl")
        
        oce.import_or_install("gensim")
        
        from gensim.models import word2vec
        
        self.model = word2vec.Word2Vec.load(model_path)
        super().__init__(log=False)

    def _convert(self, s: str) -> np.ndarray:
        mol = Chem.MolFromSmiles(s)
        sentence = MolSentence(mol2alt_sentence(mol,1))
        X = DfVec(sentences2vec([sentence], self.model, unseen='UNK')[0]).vec
        return np.array(X)

    def _convert_list(self, smiles_list: List[str], ys: List[Union[int, float, np.number]] = None) -> List[np.ndarray]:
        mols = [Chem.MolFromSmiles(s) for s in smiles_list]
        sentences = [MolSentence(mol2alt_sentence(mol, 1)) for mol in mols]
        X = [DfVec(sentence).vec for sentence in sentences2vec(sentences, self.model, unseen='UNK')]
        return np.array(X)


import copy
import io

from abc import abstractmethod
from tqdm import tqdm

import numpy as np

from .base_class import *
from .dataset import *
from .representations import *
from .external.stoned import *

import PIL

try:
    import selfies as sf
except ImportError:
    mock_imports(globals(), "sf")

from rdkit.DataStructs.cDataStructs import BulkTanimotoSimilarity, TanimotoSimilarity
from rdkit import Chem
from rdkit.Chem import AllChem

class PerturbationEngine(BaseClass):

    """PerturbationEngine is the base class for techniques which mutate or perturb a compound
    into a similar one with a small difference.

    Methods:
        get_compound_at_idx: returns a compound with a modification at a given
            atom index
        get_compound: returns a compound with a randomly chosen modification
        get_compound_list: returns a list of compounds with modifications, the list
            is meant to be comprehensive of the result of the application of an
            entire class of modifications."""

    @abstractmethod
    def get_compound_at_idx(self, mol: Chem.Mol, idx: int) -> str:
        pass

    @abstractmethod
    def get_compound(self, smiles, n=1, **kwargs) -> str:
        pass

    @abstractmethod
    def get_compound_list(self, smiles, **kwargs) -> list:
        pass

class SwapMutations(PerturbationEngine):
    """SwapMutations replaces substructures with radius r with another substructure
    with radius < r. The substructure is chosen such that it has the same outgoing
    bonds, and this set of substructures is identified through a comprehensive
    ennumeration of a large set of lead-like compounds.

    Methods:
        get_compound: returns a compound with a randomly chosen modification
        get_compound_list: returns a list of compounds with modifications, the list
            is meant to be comprehensive of the result of the application of an
            entire class of modifications"""

    @log_arguments
    def __init__(self, radius = 0, log = True):
        self.radius = radius

        if not radius in (0, 1, 2):
            raise ValueError("radius must be 0, 1, or 2")


        self.trans = dict()
        for radius in range(radius+1):
            transformation_path = download_public_file(f"swap-mutations/trans_{radius}.json")
            with open(transformation_path, "r") as f:
                self.trans.update(json.load(f))

    def get_substitution(self, m, idx, r = 1):

        # finding atoms within radius r of atom idx
        substructure_atoms = set({idx})
        for i in range(r):
            bonds = [b for a_ in substructure_atoms for b in m.GetAtomWithIdx(a_).GetBonds()]
            substructure_atoms_ = {b.GetBeginAtomIdx() for b in bonds}.union({b.GetEndAtomIdx() for b in bonds})
            substructure_atoms = substructure_atoms.union(substructure_atoms_)

        # finding atoms connecting to substructure
        surface_atoms = {b.GetEndAtomIdx() for a_ in substructure_atoms for b in m.GetAtomWithIdx(a_).GetBonds() if not b.GetEndAtomIdx() in substructure_atoms}
        surface_atoms.update({b.GetBeginAtomIdx() for a_ in substructure_atoms for b in m.GetAtomWithIdx(a_).GetBonds() if not b.GetBeginAtomIdx() in substructure_atoms})
        surface_atoms = list(surface_atoms)
        surface_l = list()
        for surface_atom in surface_atoms:
            bonds = [b.GetIdx() for b in m.GetAtomWithIdx(surface_atom).GetBonds() if b.GetEndAtomIdx() in substructure_atoms or b.GetBeginAtomIdx() in substructure_atoms]
            surface_l.append((m.GetAtomWithIdx(surface_atom).GetAtomicNum(),
                              [m.GetBondWithIdx(b).GetBondTypeAsDouble() for b in bonds]))

        tmp = [(x1, x2, x3) for x1, x2, x3 in zip([sum(surface_b) for surface_a, surface_b in surface_l], surface_atoms, surface_l)]
        tmp.sort(key = lambda x: x[0])
        surface_l = [x for _, _, x in tmp]
        surface_atoms = [x for _, x, _ in tmp]

        return surface_l, surface_atoms, substructure_atoms

    def get_entry(self, m, idx, r = 1):
        m = copy.deepcopy(m)

        # finding atoms within radius r of atom idx
        substructure_atoms = set({idx})
        for i in range(r):
            bonds = [b for a_ in substructure_atoms for b in m.GetAtomWithIdx(a_).GetBonds()]
            substructure_atoms_ = {b.GetBeginAtomIdx() for b in bonds}.union({b.GetEndAtomIdx() for b in bonds})
            substructure_atoms = substructure_atoms.union(substructure_atoms_)

        # finding atoms connecting to substructure
        surface_atoms = {b.GetEndAtomIdx() for a_ in substructure_atoms for b in m.GetAtomWithIdx(a_).GetBonds() if not b.GetEndAtomIdx() in substructure_atoms}
        surface_atoms.update({b.GetBeginAtomIdx() for a_ in substructure_atoms for b in m.GetAtomWithIdx(a_).GetBonds() if not b.GetBeginAtomIdx() in substructure_atoms})
        surface_atoms = list(surface_atoms)
        surface_l = list()
        for surface_atom in surface_atoms:
            bonds = [b.GetIdx() for b in m.GetAtomWithIdx(surface_atom).GetBonds() if b.GetEndAtomIdx() in substructure_atoms or b.GetBeginAtomIdx() in substructure_atoms]
            surface_l.append((m.GetAtomWithIdx(surface_atom).GetAtomicNum(),
                              [m.GetBondWithIdx(b).GetBondTypeAsDouble() for b in bonds]))

        tmp = [(x1, x2, x3) for x1, x2, x3 in zip([sum(surface_b) for surface_a, surface_b in surface_l], surface_atoms, surface_l)]
        tmp.sort(key = lambda x: x[0])
        surface_l = [x for _, _, x in tmp]
        surface_atoms = [x for _, x, _ in tmp]

        for i, j in enumerate(surface_atoms):
            m.GetAtomWithIdx(j).SetAtomMapNum(i+1)

        substruct = Chem.RWMol(m)

        removal = list()
        for i in range(substruct.GetNumAtoms()):
            if i not in substructure_atoms and i not in surface_atoms:
                removal.append(substruct.GetAtomWithIdx(i))

        for atom in removal:
            substruct.RemoveAtom(atom.GetIdx())
        return surface_l, substruct.GetMol()

    def stitch(self, m):
        m = Chem.RWMol(m)
        mappings = {}
        for a in m.GetAtoms():
            if a.GetAtomMapNum() != 0:
                if not a.GetAtomMapNum() in mappings:
                    mappings[a.GetAtomMapNum()] = list()
                mappings[a.GetAtomMapNum()].append(a)
        removal = []
        for k, v in mappings.items():
            i = m.AddAtom(Chem.Atom(v[0].GetAtomicNum()))
            a = m.GetAtomWithIdx(i)

            added = []
            for v_ in v:
                for b in v_.GetBonds():
                    if v_.GetIdx() == b.GetEndAtomIdx():
                        i = b.GetBeginAtomIdx()
                    else:
                        i = b.GetEndAtomIdx()
                    if not i in added:
                        m.AddBond(a.GetIdx(), i, b.GetBondType())
                        added.append(i)

                removal.append(v_)


        for remove in removal:
            m.RemoveAtom(remove.GetIdx())
        return m

    def get_compound_at_idx(self, mol, idx, **kwargs):
        ref_m = mol

        l, a, sub_a = self.get_substitution(ref_m, idx, r = self.radius)
        # if the current substructure to replace has no replacements in the database
        if str(l) not in self.trans:
            return None

        for x in sub_a:
            ref_m.GetAtomWithIdx(x).SetAtomMapNum(10000)

        for k, j in enumerate(a):
            ref_m.GetAtomWithIdx(j).SetAtomMapNum(k+1)

        sub = np.random.choice(self.trans[str(l)])
        m = Chem.CombineMols(ref_m, Chem.MolFromSmiles(sub, sanitize=False))

        m = Chem.RWMol(m)

        removal = []
        for i in range(len(m.GetAtoms())):
            if m.GetAtomWithIdx(i).GetAtomMapNum() == 10000:
                removal.append(m.GetAtomWithIdx(i))

        for remove in removal:
            m.RemoveAtom(remove.GetIdx())
        if m is None:
            errors+=1
        try:
            s1 = Chem.MolToSmiles(m)
            m = self.stitch(m)
            if not m is None:
                return Chem.MolToSmiles(m)
        except Exception as e:
            return None

    def get_compound(self, smiles, **kwargs):
        if isinstance(smiles, Chem.Mol):
            mol = smiles
            smiles = Chem.MolToSmiles(mol)
        else:
            mol = Chem.MolFromSmiles(smiles)
        idx = np.random.choice(mol.GetNumAtoms(), replace = False)
        return self.get_compound_at_idx(mol, idx, **kwargs)

    def get_compound_list(self, smiles, idx: int = None, **kwargs) -> list:
        outs = []
        ref_m = Chem.MolFromSmiles(smiles)
        if idx is None:
            space = range(len(ref_m.GetAtoms()))
        else:
            space = [idx]
        for i in tqdm(space):
            ref_m = Chem.MolFromSmiles(smiles)
            l, a, sub_a = self.get_substitution(ref_m, i, r= self.radius)

            for x in sub_a:
                ref_m.GetAtomWithIdx(x).SetAtomMapNum(10000)

            for k, j in enumerate(a):
                ref_m.GetAtomWithIdx(j).SetAtomMapNum(k+1)

            surface_atoms = [ref_m.GetAtomWithIdx(j) for j in a]
            if str(l) not in self.trans.keys():
                print("Continuing")
                continue
            for sub in self.trans[str(l)]:
                m = Chem.CombineMols(ref_m, Chem.MolFromSmiles(sub, sanitize=False))

                m = Chem.RWMol(m)

                removal = []
                for i in range(len(m.GetAtoms())):
                    if m.GetAtomWithIdx(i).GetAtomMapNum() == 10000:
                        removal.append(m.GetAtomWithIdx(i))

                for remove in removal:
                    m.RemoveAtom(remove.GetIdx())
                try:
                    m = self.stitch(m)
                    s = Chem.MolToSmiles(m)
                    m = Chem.MolFromSmiles(s)
                    if not m is None:
                        outs.append(s)
                except Exception as e:
                    pass
        return list(set(outs))

    def _save(self) -> dict:
        return super()._save()

    def _load(self, d: dict):
        return super()._load(d)

class STONEDMutations(PerturbationEngine):
    """Implements STONED-SELFIES algorithm for generating modified compounds.

    `STONED-SELFIES GitHub repository <https://github.com/aspuru-guzik-group/stoned-selfies>`_
    `Beyond Generative Models: Superfast Traversal, Optimization, Novelty, Exploration and Discovery
    (STONED) Algorithm for Molecules using SELFIES <https://doi.org/10.26434/chemrxiv.13383266.v2>`_

    Methods:
        get_compound_at_idx: returns a compound with a randomly chosen modification at `idx`
        get_compound: returns a compound with a randomly chosen modification
        get_compound_list: returns a list of `num_samples` compounds with randomly chosen modifications
    """

    @log_arguments
    def __init__(self, mutations: int = 1, log = True):
        self.mutations = mutations

    def get_compound_at_idx(self, mol: Chem.Mol, idx: int, **kwargs) -> str:
        smiles = Chem.MolToSmiles(mol)
        selfie = sf.encoder(smiles)
        selfie_chars = get_selfie_chars(selfie)
        max_molecules_len = len(selfie_chars) + self.mutations
        selfie_mutated, _ = mutate_selfie(selfie, max_molecules_len, index = idx)
        smiles_back = sf.decoder(selfie_mutated)
        return smiles_back

    def get_compound(self, smiles: str, **kwargs) -> str:
        ref_m = Chem.MolFromSmiles(smiles)
        randomized_smiles_ordering = randomize_smiles(ref_m)
        selfie = sf.encoder(randomized_smiles_ordering)
        selfie_mutated = get_mutated_SELFIES([selfie], num_mutations = self.mutations)[0]
        smiles_back = sf.decoder(selfie_mutated)
        return smiles_back

    def get_compound_list(self, smiles: str, num_samples: int = 1000, **kwargs) -> list:
        ref_m = Chem.MolFromSmiles(smiles)
        randomized_smile_orderings = [randomize_smiles(ref_m) for _ in range(num_samples)]
        selfies_ls = [sf.encoder(x) for x in randomized_smile_orderings]
        selfies_mut = get_mutated_SELFIES(selfies_ls.copy(), num_mutations = self.mutations)
        smiles_back = [sf.decoder(x) for x in selfies_mut]
        return list(set(smiles_back))

    def _save(self) -> dict:
        return super()._save()

    def _load(self, d: dict):
        return super()._load(d)

class CounterfactualEngine(BaseClass):
    """Generates counterfactual compounds based on:

    `exmol GitHub repository <https://github.com/ur-whitelab/exmol>`_
    `Model agnostic generation of counterfactual explanations for molecules <http://dx.doi.org/10.1039/D1SC05259D>`_
    """

    @log_arguments
    def __init__(self, model: BaseModel, perturbation_engine: PerturbationEngine = "default"):
        self.model = model
        if perturbation_engine == "default":
            self.perturbation_engine = SwapMutations(radius=1)
        else:
            self.perturbation_engine = perturbation_engine
        self.samples = []
        self.cfs = []

    def generate_samples(self, smiles: str) -> None:
        """Generates candidate counterfactuals and stores them in `self.samples` as a list of dictionaries.

        Args:
            smiles: SMILES string of the target prediction
        """
        perturbed_smiles = [smiles] + self.perturbation_engine.get_compound_list(smiles)
        filtered_smiles = []
        seen = set()
        for smi in tqdm(perturbed_smiles):
            mol, smi_canon, conversion_successful = sanitize_smiles(smi)
            if conversion_successful and smi_canon not in seen:
                filtered_smiles.append(smi_canon)
                seen.add(smi_canon)

        ref_m = Chem.MolFromSmiles(smiles)
        filtered_mols = [Chem.MolFromSmiles(s) for s in filtered_smiles]

        ref_fp = AllChem.GetMorganFingerprint(ref_m, 2)
        fps = [AllChem.GetMorganFingerprint(m, 2) for m in filtered_mols]
        scores = BulkTanimotoSimilarity(ref_fp, fps)

        values = self.model.predict(filtered_smiles)

        dmat = 1 - np.array([BulkTanimotoSimilarity(fp, fps) for fp in tqdm(fps)])

        from sklearn.cluster import DBSCAN
        from sklearn.decomposition import PCA
        coords = PCA(n_components = 2).fit_transform(dmat)
        clustering = DBSCAN(eps=0.15, min_samples=5).fit(coords)

        samples = {
            'SMILES': filtered_smiles,
            'Similarity': scores,
            'Value': values,
            'Coordinate': list(coords),
            'Cluster': clustering.labels_,
            'Original': [True] + [False for _ in filtered_smiles[1:]]
        }
        self.samples = pd.DataFrame(samples).to_dict('records')

    def get_samples(self) -> pd.DataFrame:
        """Returns candidate counterfactuals as a pandas dataframe.

        Returns:
            pandas dataframe of candidate counterfactuals
        """
        return pd.DataFrame(self.samples)

    def generate_cfs(self, delta: Union[int, float, Tuple] = (-1, 1), n: int = 4) -> None:
        """Generates counterfactuals and stores them in `self.cfs` as a list of dictionaries.

        Args:
            delta: margin defining counterfactuals for regression models
            n: number of counterfactuals
        """
        base = self.samples[0]['Value']

        if self.model.setting == "classification":
            cfs = self._select_cfs(lambda s: s['Value'] != base, n)
            self.cfs = self.samples[:1] + cfs
        else:
            if type(delta) in (float, int):
                delta = (-delta, delta)
            lcfs = self._select_cfs(lambda s: s['Value'] - delta[0] < base, n // 2)
            hcfs = self._select_cfs(lambda s: s['Value'] - delta[1] > base, n // 2)
            self.cfs = self.samples[:1] + lcfs + hcfs

    def get_cfs(self) -> pd.DataFrame:
        """Returns counterfactuals as a pandas dataframe.

        Returns:
            pandas dataframe of counterfactuals
        """
        return pd.DataFrame(self.cfs)

    def _select_cfs(self, fn: Callable, n: int) -> list:
        """Selects counterfactuals from `self.samples` using a validity function.

        Args:
            fn: boolean function for identifying counterfactuals
            n: maximum number of counterfactuals
        Returns:
            cfs: list of dictionary of counterfactuals
        """
        cfs = []

        def cluster_score(s, i):
            """Score for clustering for a given model.

            Parameters:
                s: sample
                i: index of sample

            Returns:
                score: score for clustering
            """
            return (s['Cluster'] == i) * fn(s) * s['Similarity']

        clusters = {s['Cluster'] for s in self.samples[1:]}
        for i in clusters:
            candidate = max(self.samples[1:], key=lambda s: cluster_score(s, i))
            if cluster_score(candidate, i):
                cfs.append(candidate)

        cfs = sorted(cfs, key=lambda s: s['Similarity'], reverse=True)[:n]
        return cfs

    def _save(self) -> dict:
        return super()._save()

    def _load(self, d: dict):
        return super()._load(d)

def model_molecule_sensitivity(model: BaseModel, smiles: str, perturbation_engine: PerturbationEngine = "default", n: int = 30) -> Chem.Mol:

    """
    Calculates the sensitivity of a model to perturbations in on each of a molecule's atoms,
    outputting a rdkit molecule, with sensitivity as an atom property.

    Parameters:
        model: model to be used for sensitivity calculation
        smiles: SMILES string of the target prediction
        perturbation_engine: perturbation engine to be used for sensitivity calculation
        n: number of perturbations to be used for sensitivity calculation

    Returns:
        rdkit molecule with sensitivity as an atom property
    """

    if perturbation_engine == "default":
        perturbation_engine = SwapMutations(radius=1)

    mol = Chem.MolFromSmiles(smiles)

    idxs = []

    for a in tqdm(mol.GetAtoms()):
        idx = a.GetIdx()
        smiles = [Chem.MolToSmiles(mol)]

        count = 1
        while count < n:
            perturbed_smiles = perturbation_engine.get_compound_at_idx(smiles[0], idx)
            try:
                mol2 = Chem.MolFromSmiles(perturbed_smiles)
                if not mol2 is None:
                    smiles.append(perturbed_smiles)
                    idxs.append(idx)
                    count += 1
            except:
                pass
        predictions = model.predict(smiles)
        v = np.std(predictions)
        a.SetProp("sensitivity", str(v))
        a.SetProp("atomNote", str(np.format_float_scientific(v, precision=2)))
    return mol


import base64
import hashlib
import inspect
import json
import os
import pickle
from abc import ABC, abstractmethod
from typing import Callable, Union
import gcsfs
import subprocess
import pyrebase  # Default pyrebase is pyrebase3 which won't work. Need to install pyrebase4 (pip install pyrebase4)
from google.cloud.firestore import Client
from google.oauth2.credentials import Credentials
import sys
from unittest.mock import MagicMock

import olorenchemengine

sys.modules["olorenautoml"] = olorenchemengine # important for backwards compatibility of some models

def mock_imports(g, *args):
    for arg in args:
        g[arg] = MagicMock()

def all_subclasses(cls):
    """ Helper function to return all subclasses of class"""
    return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in all_subclasses(c)])

class OASConnector:
    """ Class which links oce to OAS and can move data between them using Firestore and an API"""

    def __init__(self) -> None:
        self.storage = None

    def authenticate(self):
        import json

        import requests
        from requests.exceptions import HTTPError

        FIREBASE_REST_API = "https://identitytoolkit.googleapis.com/v1/accounts"
        FIREBASE_API_KEY = "AIzaSyCSaFE5-qxy5-HAr-oqUd5del4bv5dqJ0Q"

        # Use the google verify password REST api to authenticate and generate user tokens
        def log_in_first_time(token):
            request_url = f"{FIREBASE_REST_API}:signInWithCustomToken?key={FIREBASE_API_KEY}"
            headers = {"content-type": "application/json; charset=UTF-8"}
            data = json.dumps({"token": token, "returnSecureToken": True})
            resp = requests.post(request_url, headers=headers, data=data)

            # Check for errors
            try:
                resp.raise_for_status()
            except HTTPError as e:
                raise HTTPError(e, resp.text)

            return resp.json()

        def get_id_token_from_refresh(refresh_token):
            request_url = f"https://securetoken.googleapis.com/v1/token?key={FIREBASE_API_KEY}"
            headers = {"content-type": "application/json; charset=UTF-8"}
            data = json.dumps({"refresh_token": refresh_token, "grant_type": "refresh_token"})
            resp = requests.post(request_url, headers=headers, data=data)
            try:
                resp.raise_for_status()
            except HTTPError as e:
                raise HTTPError(e, resp.text)

            return resp.json()

        from os.path import expanduser

        home = expanduser("~")
        PATH_TO_REFRESH_TOKEN = f"{home}/.oce/firebase_refresh_token.json"

        if not os.path.exists(PATH_TO_REFRESH_TOKEN):
            print("Navigate to https://oas.oloren.ai/auth to get auth token.")
            response = log_in_first_time(input("Paste token here: "))
            if not os.path.exists(f"{home}/.oce"):
                os.makedirs(f"{home}/.oce")
            with open(PATH_TO_REFRESH_TOKEN, "w") as f:
                f.write(json.dumps({"refreshToken": response["refreshToken"]}))

        with open(PATH_TO_REFRESH_TOKEN, "r") as f:
            refresh_token = json.loads(f.read())["refreshToken"]
            response = get_id_token_from_refresh(refresh_token)
            response["idToken"] = response["id_token"]
            response["refreshToken"] = response["refresh_token"]
            response["uid"] = response["user_id"]

        with open(PATH_TO_REFRESH_TOKEN, "w") as f:
            f.write(json.dumps({"refreshToken": response["refreshToken"]}))

        creds = Credentials(response["idToken"], response["refreshToken"])

        """Storage"""

        config = {
            "apiKey": "AIzaSyCSaFE5-qxy5-HAr-oqUd5del4bv5dqJ0Q",
            "authDomain": "oloren-ai.firebaseapp.com",
            "databaseURL": "https://oloren-ai-default-rtdb.firebaseio.com",
            "projectId": "oloren-ai",
            "storageBucket": "oloren-ai.appspot.com",
            "messagingSenderId": "602366687071",
            "appId": "1:602366687071:web:f35531f7142084b86a28ea",
            "measurementId": "G-MVNK4DPQ60",
        }
        firebase = pyrebase.initialize_app(config)

        auth = firebase.auth()
        user = auth.refresh(response["refreshToken"])

        self.storage = firebase.storage()
        self.logging_db = Client("oloren-ai", creds)
        self.uid = response["uid"]
        self.uid_token = response["idToken"]

    def upload_vis(self, visualization):
        self.authenticate()

        out = self.logging_db.collection("visualizations").add(
            {
                "uid": self.uid,
                "status": "Visualization rendered",
                "data_url": visualization.render_data_url(),
            }
        )

        print(f"Uploaded visualization to https://oas.oloren.ai/vis?vid={out[1].id}")

        return out


    def upload_model(self, model, model_name):
        self.authenticate()

        import tempfile

        with tempfile.NamedTemporaryFile() as tmp:
            pickle.dump(saves(model), tmp)
            tmp.flush()
            self.storage.child(self.uid + "/models/" + model_name + ".pkl").put(tmp.name, self.uid_token)

        out = self.logging_db.collection("models").add(
            {
                "uid": self.uid,
                "did": None,
                "source": "oce",
                "data": {
                    "file_path": "gs://oloren-ai.appspot.com/" + self.uid + "/models/" + model_name + ".pkl",
                    "name": model_name,
                    "description": pretty_params_str(model),
                    "parameters": json.dumps(parameterize(model)),
                    "metrics": None,
                    "status": "Runnable",
                },
            }
        )

        return out

oas_connector = OASConnector()

ignored_kwargs = ["map_location", "num_workers"]

def download_public_file(path, redownload=False):
    """Download a public file from Oloren's Public Storage, and returns the contents.

    @param path: The path to the file to read.
    @param redownload: Whether to redownload the file if it already exists.
    """

    local_path = os.path.join(os.path.expanduser("~"), ".oce", path)

    if os.path.exists(local_path) and not redownload: return local_path

    if not os.path.exists(os.path.dirname(local_path)):
        os.makedirs(os.path.dirname(local_path))

    print(f"Downloading {path}...")
    fs = gcsfs.GCSFileSystem()
    with fs.open(f'gs://oloren-public-data/{path}', 'rb') as f:
        with open(local_path, 'wb') as out:
            out.write(f.read())

    return local_path


def get_default_args(func):
    signature = inspect.signature(func)
    return {k: v.default for k, v in signature.parameters.items() if v.default is not inspect.Parameter.empty}

def log_arguments(func: Callable[..., None]) -> Callable[..., None]:
    """
    log_arguments is a decorator which logs the arguments of a BaseClass constructor to instance variables for use in
        model parameterization.

    Args:
        func (function): a __init__(self, *args, **kwargs) function of a baseclass.

    Returns:
        wrapper (function): the same __init__ function with arguments saved to instance variables.
    """

    def wrapper(self, *args, **kwargs):

        if "log" in kwargs.keys() and not kwargs["log"]:
            pass
        else:
            kwds = get_default_args(func)
            for k, v in kwds.items():
                if not k in kwargs.keys():
                    kwargs.update({k: v})
            self.args = args
            self.kwargs = {k: v for k, v in kwargs.items() if not k in ignored_kwargs}
        return func(self, *args, **kwargs)

    wrapper.__wrapped__ = func

    return wrapper

class BaseClass(ABC):
    """BaseClass is the base class for all models.

    All classes in Oloren ChemEngine should inherit from BaseClass to enable for universal saving and loading of both
    parameters and internal state. This requires the implementation of abstract methods _save and _load.

    Methods:
        Registry: returns a dictionary mapping the name of a class to the class itself for all subclasses of the class.
        _save: saves an instance of a BaseClass to a dictionary (abstract method to be implmented by subclasses)
        _load: loads an instance of a BaseClass from a dictionary (abstract method to be implmented by subclasses)
    """

    @classmethod
    def Registry(cls):
        """Registry is a recursive method to create a dictionary of all subclasses of BaseClass, with the key being the name of the subclass and the value being the subclass itself."""
        d = {cls.__name__: cls}
        scs = cls.__subclasses__()
        if len(scs) > 0:
            for sc in scs:
                u = sc.Registry()
                d = {**d, **u}
        return d

    @classmethod
    def AllInstances(cls):
        """AllTypes returns a list of all standard instances of all subclasses of BaseClass.

        Standard instances means that all required parameters for instantiation of the
        subclasses are set with canonical values."""
        if not hasattr(cls, "__abstractmethods__") or len(cls.__abstractmethods__) == 0:
            try:
                return [cls()] + [o for sc in cls.__subclasses__() for o in sc.AllInstances()]
            except:
                return [o for sc in cls.__subclasses__() for o in sc.AllInstances()]
        else:
            return [o for sc in cls.__subclasses__() for o in sc.AllInstances()]

    @classmethod
    def Opt(cls, *args, **kwargs):
        return {
            **{"BC_class_name": cls.__name__},
            **{"args": args},
            **{"kwargs": kwargs},
        }

    @abstractmethod
    def _save(self) -> dict:
        """_save is an abstract method that must be implemented by all subclasses. It should return a dictionary of variables which can passed to an instance of a model via _load to completely recreate a model.

        Returns:
            dict: a dictionary of variables which can be passed to _load to recreate the model.
        """
        return ""

    @abstractmethod
    def _load(self, d: dict):
        """_load is an abstract method that must be implemented by all subclasses. It should take an instance save by _save and recreate the model."""
        pass

    def copy(self):
        obj_copy = create_BC(parameterize(self))
        obj_copy._load(self._save())
        return obj_copy


def parameterize(object: Union[BaseClass, list, int, float, str, None]) -> dict:
    """parameterize is a recursive method which creates a dictionary of all arguments necessary to instantiate a BaseClass object.

    Note that only objects which are instances of subclasses of BaseClass can be parameterized, other supported objects are to enable to recursive use of parameterize but cannot themselves be parameterized.

    Args:
        object (Union[BaseClass, list, int, float, str, None]): parameterize is a recursive method which creates a dictionary of all arguments necessary to instantiate a BaseClass object.

    Raises:
        ValueError: Object is not of type that can be parameterized

    Returns:
        dict: dictionary of parameters necessary to instantiate the object.
    """
    if issubclass(type(object), BaseClass):
        return {
            **{"BC_class_name": type(object).__name__},
            **{"args": [parameterize(arg) for arg in object.args]},
            **{"kwargs": {k: parameterize(v) for k, v in object.kwargs.items()}},
        }
    elif (
        object is None
        or issubclass(type(object), int)
        or issubclass(type(object), float)
        or issubclass(type(object), str)
    ):
        return object
    elif issubclass(type(object), list):
        return [parameterize(x) for x in object]
    else:
        print(object)
        raise ValueError


def model_name_from_params(param_dict: dict) -> str:
    """model_name_from_params creates a unique name for a model based on the parameters passed to it.

    Args:
        param_dict (dict): dictionary of parameters returned by parameterize neccessary to instantiate the model (note this is different from the instance save)

    Returns:
        str: the model name consisting of the the model class name with a hash of the parameters
    """
    return (
        param_dict["BC_class_name"]
        + " "
        + base64.urlsafe_b64encode(hashlib.md5(str(param_dict).encode("utf-8")).digest())[:8].decode("utf-8")
    )


def model_name_from_model(model: BaseClass) -> str:
    """model_name_from_model creates a unique name for a model.

    Args:
        model (BaseClass): the model to be named

    Returns:
        str: the model name consisting of the the model class name with a hash of the parameters
    """
    return model_name_from_params(parameterize(model))


def saves(object: Union[BaseClass, dict, list, int, float, str, None]) -> dict:
    """saves is a method which saves BaseClass object, which can be recovered via loads.

    Args:
        object (Union[BaseClass, dict, list, int, float, str, None]): the object to be saved

    Returns:
        dict: a dictionary which can be passed to loads to recreate the object
    """
    if issubclass(type(object), BaseClass):
        return {
            **{"BC_class_name": type(object).__name__},
            **{"instance_save": object._save()},
            **{"args": [saves(arg) for arg in object.args]},
            **{"kwargs": {k: saves(v) for k, v in object.kwargs.items()}},
        }
    elif (
        issubclass(type(object), int)
        or issubclass(type(object), float)
        or issubclass(type(object), str)
        or object is None
    ):
        return object
    elif issubclass(type(object), list):
        return [saves(x) for x in object]
    elif issubclass(type(object), dict):
        return {k: saves(v) for k, v in object.items()}
    else:
        print(object)
        raise ValueError


def create_BC(d: dict) -> BaseClass:
    """create_BC is a method which creates a BaseClass object from a dictionary of parameters.

    Note the instances variables of the object are not specified.

    Args:
        d (dict): a dictionary of parameters returned by parameterize

    Returns:
        BaseClass: the object created from the parameters
    """
    if isinstance(d, str):
        d = d.replace("'", '"').replace("True", "true").replace("False", "false").replace("None", "null")
        d = json.loads(d)
    args = []
    if "args" in d.keys():
        for arg in d["args"]:
            if isinstance(arg, dict) and "BC_class_name" in arg.keys():
                args.append(create_BC(arg))
            else:
                if isinstance(arg, list):
                    arg = [create_BC(x) if isinstance(x, dict) and "BC_class_name" in x.keys() else x for x in arg]
                args.append(arg)

    kwargs = {}
    if "kwargs" in d.keys():
        for k, v in d["kwargs"].items():
            if isinstance(v, dict) and "BC_class_name" in v.keys():
                kwargs[k] = create_BC(v)
            else:
                kwargs[k] = v

    return BaseClass.Registry()[d["BC_class_name"]](*args, **kwargs)


def loads(d: dict) -> BaseClass:
    """loads is a method which recreates a BaseClass object from a save.

    Args:
        d (dict): the dictionary returned by saves which saves the state of a BaseClass object

    Returns:
        BaseClass: the recreated object
    """

    import olorenchemengine as olorenautoml # for backwards compatibility

    args = []
    for arg in d["args"]:
        if isinstance(arg, dict) and "BC_class_name" in arg.keys():
            args.append(loads(arg))
        else:
            if isinstance(arg, list):
                arg = [loads(x) if isinstance(x, dict) and "BC_class_name" in x.keys() else x for x in arg]
            args.append(arg)

    kwargs = {k: loads(v) if isinstance(v, dict) and "BC_class_name" in v.keys() else v for k, v in d["kwargs"].items()}
    bc = BaseClass.Registry()[d["BC_class_name"]](*args, **kwargs)
    bc._load(d["instance_save"])
    return bc


def save(model: BaseClass, fname: str):
    """saves a BaseClass object to a file

    Args:
        model (BaseClass): the object to be saved
        fname (str): the file name to save the model to
    """
    save_dict = saves(model)
    with open(fname, "wb+") as f:
        pickle.dump(save_dict, f)


def load(fname: str) -> BaseClass:
    """loads a BaseClass from a file

    Args:
        fname (str): name of the file to load the object from

    Returns:
        BaseClass: the BaseClass object which as been recreated from the file
    """
    with open(fname, "rb") as f:
        d = pickle.load(f)
    return loads(d)


def pretty_params(base: Union[BaseClass, dict]) -> dict:
    """ Returns a dictionary of the parameters of the passed BaseClass object, formatted such that they are in a
    human readable format, with the names of the arguments included."""
    if isinstance(base, dict):
        base = loads(base)

    if issubclass(type(base), BaseClass):
        args = [arg for arg in base.args]
        kwargs = {k: v for k, v in base.kwargs.items()}
        base_object_parameters = list(inspect.signature(base.__init__).parameters.keys())

        for kwarg in kwargs:
            if kwarg in base_object_parameters:
                base_object_parameters.remove(kwarg)

        labeled_args = {k: v for k, v in zip(base_object_parameters, args)}

        fully_labeled_args = {k: pretty_params(v) for k, v in labeled_args.items()}
        fully_labeled_kwargs = {k: pretty_params(v) for k, v in kwargs.items()}

        return {**{"BC_Class_name": type(base).__name__}, **fully_labeled_args, **fully_labeled_kwargs}
    elif issubclass(type(base), int) or issubclass(type(base), float) or issubclass(type(base), str) or base is None:
        return base
    elif issubclass(type(base), list):
        return [pretty_params(x) for x in base]
    elif issubclass(type(base), dict):
        return {k: pretty_params(v) for k, v in base.items()}
    else:
        print(type(base))
        raise ValueError


def pretty_params_str(base: Union[BaseClass, dict]) -> str:
    """ Returns a string of the parameters of the passed BaseClass object, formatted such that they are in a human
    readable format"""
    return json.dumps(pretty_params(base), indent=4)


def json_params_str(base: Union[BaseClass, dict]) -> str:
    """ Returns a json string of the parameters of the passed BaseClass object so that the model parameter dictionary can
    be reconstructed with json.load(params_str)"""
    return (
        json.dumps(pretty_params(base))
        .replace("'", '"')
        .replace("True", "true")
        .replace("False", "false")
        .replace("None", "null")
    )

def install_with_permission(package_name: str):
    inp = input(f"The required package {package_name} is not installed. Do you want to install it? [y/N]? ")
    if inp.lower() == "y":
        subprocess.check_call([sys.executable, "-m", "pip", "install", package_name])
    else:
        print(f"Stopping program. You can install the package manually with: \n >> pip install {package_name}")
        os._exit(1)

def import_or_install(package_name: str, statement: str = None, scope: dict = None):
    if scope is None:
        scope = globals()
    if statement is None:
        statement = f"import {package_name}"
    try:
        exec(statement, scope)
    except ImportError:
        install_with_permission(package_name)
    finally:
        exec(statement, scope)


"""piCalculax is a library for isoelectric point prediction for chemically modified peptides.

The original library has been adapted to be incorporated into Oloren ChemEngine

For more information visit the github repository: https://github.com/EBjerrum/pICalculax
"""

from rdkit import Chem
from rdkit.Chem import AllChem
from numpy import unique, argmin
from numpy import abs as np_abs

# RuleTables
from rdkit import Chem

# Format of ruletable tuple (Name, SMARTS, pKa (evt. as list), type (evt.as list) )
# Type is 0 = acidic (charge below pKa)
# Type 1 is basic (charge below pKa)

# (Grimsley 2009) with extensions
# Grimsley, G. R.; Scholtz, J. M. & Pace, C. N. (2009), 'A summary of the measured pK values of the ionizable groups in folded proteins.', Protein science : a publication of the Protein Society 18, 247--251.

# Default pKa tables defined at the end of the file

pkatable_grimsley = [
    ('Tetramethylrhodamine #3',
     Chem.MolFromSmiles('CN(C)C1=C-C=C2-C(C4=C(C(=O)O)C=CC=C4)=C3-C=C-C(=[N+](C)C)C=C3OC2=C1'), [-10, 1.19, 3.59, 10],
     [1, 1, 0, 1]),
    ('6-hydroxy-3H-xanthen-3-one', Chem.MolFromSmarts('c2c3ccc(O)cc3oc3cc(=O)ccc2-3'), [2.94, 6.56], [1, 0]),
    ('Anilin', Chem.MolFromSmarts('[$([NH2]c1ccccc1)]'), 4.87, 1),
    ('N-acetyl Hydrazine', Chem.MolFromSmarts('[$([NH2]NC(=O))]'), 2.81, 1),  # Reaxys
    ('Hydrazine', Chem.MolFromSmarts('[$([NH2]N)]'), 8.1, 1),  # Wikiped
    ('N-acetyl Piperazine', Chem.MolFromSmarts('[N+&!H0,NX3&H0;$(N1(C)CCN(C(=O)C)CC1)]'), 7.1, 1),  # Reaxys
    ('Piperazine', Chem.MolFromSmarts('C1CNCCN1'), 9.73, 1),  # Wikiped
    ('Pyridine', Chem.MolFromSmarts('n1ccccc1'), 5.25, 1),  # Wikiped
    ('Sulfonate', Chem.MolFromSmarts('[$([OH]S(=O)=O)]'), -1.5, 0),  # Guess
    ('Phosphonate', Chem.MolFromSmarts('[OH]P([OH])(=O)[#6]'), [1.5, 7.0], [0, -1]),  # Guess
    ('7H-Pyrimido[4,5-b][1,4]oxazine-2,4-diamine', Chem.MolFromSmarts('C1C=NC2=CN=C(N=C2O1)N'), 3.6, 1),  # PKA Guessed
    ('Tyrosine', Chem.MolFromSmarts('[$([O-,OH]c1ccccc1)]'), 10.3, 0),
    ('Histidine', Chem.MolFromSmarts(
        '[CH2X4][#6X3]1:[$([#7X3H+,#7X2H0+0]:[#6X3H]:[#7X3H]),$([#7X3H])]:[#6X3H]:[$([#7X3H+,#7X2H0+0]:[#6X3H]:[#7X3H]),$([#7X3H])]:[#6X3H]1'),
     6.6, 1),
    ('Histidine1', Chem.MolFromSmarts('[$(cC)]1cnc[nH]1'), 6.6, 1),
    ('Histidine2', Chem.MolFromSmarts('[$(cC)]1c[nH]cn1'), 6.6, 1),
    ('Glutamic Acid', Chem.MolFromSmarts('[$([CX3][CD2][CD2]C(N)C(=O))](=O)[OX1H0-,OX2H1]'), 4.2, 0),
    ('Cysteine', Chem.MolFromSmarts('[$([S-,SH][CD2]C(N)C(=O))]'), 6.8, 0),
    ('Condensed Cysteine', Chem.MolFromSmarts('[$([SH][#175])]'), 6.8, 0),  # Sidechain treated as normal atom
    ('Condensed Cysteine', Chem.MolFromSmarts('[$([SH][#178])]'), 6.8, 0),  # Sidechain treated as normal atom
    # ('Aspartic Acid', Chem.MolFromSmarts('[$([CX3][CD2])](=O)[OX1H0-,OX2H1]'),3.90,0),
    ('Aspartic Acid v2', Chem.MolFromSmarts('[$([CX3][CD2]C(N)C(=O))](=O)[OX1H0-,OX2H1]'), 3.5, 0),
    ('Arginine', Chem.MolFromSmarts('[$(N[CD2][CD2])]C(=N)N'), 12.0, 1),
    ('Lysine', Chem.MolFromSmarts('[$([ND1][CD2][CD2][CD2][CD2]C(N)C(=O))]'), 10.5, 1),
    ('End Carboxylate (avg)', Chem.MolFromSmarts('[$([CX3]CN)](=O)[OX1H0-,OX2H1]'), 3.3, 0),
    # ('End Amine (avg)',Chem.MolFromSmarts('[$([NX3;H2,H1;!$(NC=O)]CC=O)]'),9.15,1),
    ('End Amine (avg) PseudoAtom safe', Chem.MolFromSmarts(
        '[$([NX3;H2,H1;!$(NC=O);!$(N[#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193])]CC=O)]'),
     7.7, 1),
    ('Acrydine', Chem.MolFromSmarts('c1c2ccccc2nc2ccccc21'), 6.15, 1),
    ('Generic Carboxylic Acid', Chem.MolFromSmarts('[CX3](=O)[OX1H0-,OX2H1]'), 4.75, 0),
    (
    'Generic tert. Amine, not amide', Chem.MolFromSmarts('[N+&!H0&NX3,NX3;$(N(C)(C)C);!$(NC=[!#6]);!$(NC#[!#6])]'), 9.8,
    1),
    ('Generic Amine, not amide', Chem.MolFromSmarts(
        '[N+,NX3;!H0;!$(NC=[!#6]);!$(NC#[!#6])!$(N[#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193])]'),
     10., 1),
]

condensed_grimsley = [
    ('Condensed Tyrosine', Chem.MolFromSmarts('[#189]'), 10.3, 0),
    ('Condensed Histidine', Chem.MolFromSmarts('[#179]'), 6.6, 1),
    ('Condensed Glutamic Acid', Chem.MolFromSmarts('[#177]'), 4.2, 0),
    ('Condensed Aspartic Acid', Chem.MolFromSmarts('[#174]'), 3.5, 0),
    ('Condensed Arginine', Chem.MolFromSmarts('[#172]'), 12.0, 1),
    ('Condensed Lysine', Chem.MolFromSmarts('[#182]'), 10.5, 1),
    # ('Condensed AA end amino subst.', Chem.MolFromSmarts('[#0D1]'),9.15,1),
    # Can it somehow match up the wrong end? with a protection group??
    ('Condensed AA end amino', Chem.MolFromSmarts(
        '[#171,#172,#173,#174,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193;D1]'),
     7.7, 1),
    # ('Condensed AA end Carboxylic Subst.', Chem.MolFromSmarts('[$([#0][OH])]'),2.15,0), #All acidic/basic should be replaced with glycine so they can be removed??
    ('Condensed cys end amino', Chem.MolFromSmarts('[#175;D2]'), 7.7, 1),
    # ('Condensed AA end Carboxylic Subst.', Chem.MolFromSmarts('[$([#0][OH])]'),2.15,0), #All acidic/basic should be replaced with glycine so they can be removed??
    ('Condensed AA end Carboxylic', Chem.MolFromSmarts(
        '[$([#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193][OH])]'),
     3.3, 0)
]

# IPC_peptide as refered in KOZLOWSKI, Lukasz P. "IPC - Isoelectric Point Calculator.". Biol. Direct. 2016, vol 11, p. 55.
pkatable_ipc = [
    ('Tetramethylrhodamine #3',
     Chem.MolFromSmiles('CN(C)C1=C-C=C2-C(C4=C(C(=O)O)C=CC=C4)=C3-C=C-C(=[N+](C)C)C=C3OC2=C1'), [-10, 1.19, 3.59, 10],
     [1, 1, 0, 1]),
    # Guess DOI: 10.1039/C0OB01045F
    ('6-hydroxy-3H-xanthen-3-one', Chem.MolFromSmarts('c2c3ccc(O)cc3oc3cc(=O)ccc2-3'), [2.94, 6.56], [1, 0]),
    ('Anilin', Chem.MolFromSmarts('[$([NH2]c1ccccc1)]'), 4.87, 1),
    ('N-acetyl Hydrazine', Chem.MolFromSmarts('[$([NH2]NC(=O))]'), 2.81, 1),  # Reaxys
    ('Hydrazine', Chem.MolFromSmarts('[$([NH2]N)]'), 8.1, 1),  # Wikiped
    ('N-acetyl Piperazine', Chem.MolFromSmarts('[N+&!H0,NX3&H0;$(N1(C)CCN(C(=O)C)CC1)]'), 7.1, 1),  # Reaxys
    ('Piperazine', Chem.MolFromSmarts('C1CNCCN1'), 9.73, 1),  # Wikiped
    ('Pyridine', Chem.MolFromSmarts('n1ccccc1'), 5.25, 1),  # Wikiped
    ('Sulfonate', Chem.MolFromSmarts('[$([OH]S(=O)=O)]'), -1.5, 0),  # Guess
    ('Phosphonate', Chem.MolFromSmarts('[OH]P([OH])(=O)[#6]'), [1.5, 7.0], [0, -1]),  # Guess
    ('7H-Pyrimido[4,5-b][1,4]oxazine-2,4-diamine', Chem.MolFromSmarts('C1C=NC2=CN=C(N=C2O1)N'), 3.6, 1),  # PKA Guessed
    ('Tyrosine', Chem.MolFromSmarts('[$([O-,OH]c1ccccc1)]'), 10.071, 0),
    ('Histidine', Chem.MolFromSmarts(
        '[CH2X4][#6X3]1:[$([#7X3H+,#7X2H0+0]:[#6X3H]:[#7X3H]),$([#7X3H])]:[#6X3H]:[$([#7X3H+,#7X2H0+0]:[#6X3H]:[#7X3H]),$([#7X3H])]:[#6X3H]1'),
     6.018, 1),
    ('Histidine1', Chem.MolFromSmarts('[$(cC)]1cnc[nH]1'), 6.018, 1),
    ('Histidine2', Chem.MolFromSmarts('[$(cC)]1c[nH]cn1'), 6.018, 1),
    ('Glutamic Acid', Chem.MolFromSmarts('[$([CX3][CD2][CD2]C(N)C(=O))](=O)[OX1H0-,OX2H1]'), 4.317, 0),
    ('Cysteine', Chem.MolFromSmarts('[$([S-,SH][CD2]C(N)C(=O))]'), 8.297, 0),
    ('Condensed Cysteine', Chem.MolFromSmarts('[$([SH][#175])]'), 8.297, 0),  # Sidechain treated as normal atom
    ('Condensed Cysteine', Chem.MolFromSmarts('[$([SH][#178])]'), 8.297, 0),  # Sidechain treated as normal atom
    # ('Aspartic Acid', Chem.MolFromSmarts('[$([CX3][CD2])](=O)[OX1H0-,OX2H1]'),3.90,0),
    ('Aspartic Acid v2', Chem.MolFromSmarts('[$([CX3][CD2]C(N)C(=O))](=O)[OX1H0-,OX2H1]'), 3.887, 0),
    ('Arginine', Chem.MolFromSmarts('[$(N[CD2][CD2])]C(=N)N'), 12.503, 1),
    ('Lysine', Chem.MolFromSmarts('[$([ND1][CD2][CD2][CD2][CD2]C(N)C(=O))]'), 10.517, 1),
    ('End Carboxylate (avg)', Chem.MolFromSmarts('[$([CX3]CN)](=O)[OX1H0-,OX2H1]'), 2.383, 0),
    # ('End Amine (avg)',Chem.MolFromSmarts('[$([NX3;H2,H1;!$(NC=O)]CC=O)]'),9.15,1),
    ('End Amine (avg) PseudoAtom safe', Chem.MolFromSmarts(
        '[$([NX3;H2,H1;!$(NC=O);!$(N[#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193])]CC=O)]'),
     9.564, 1),
    ('Acrydine', Chem.MolFromSmarts('c1c2ccccc2nc2ccccc21'), 6.15, 1),
    ('Generic Carboxylic Acid', Chem.MolFromSmarts('[CX3](=O)[OX1H0-,OX2H1]'), 4.75, 0),
    (
    'Generic tert. Amine, not amide', Chem.MolFromSmarts('[N+&!H0&NX3,NX3;$(N(C)(C)C);!$(NC=[!#6]);!$(NC#[!#6])]'), 9.8,
    1),
    ('Generic Amine, not amide', Chem.MolFromSmarts(
        '[N+,NX3;!H0;!$(NC=[!#6]);!$(NC#[!#6])!$(N[#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193])]'),
     10., 1),
]

condensed_ipc = [
    ('Condensed Tyrosine', Chem.MolFromSmarts('[#189]'), 10.071, 0),
    ('Condensed Histidine', Chem.MolFromSmarts('[#179]'), 6.018, 1),
    ('Condensed Glutamic Acid', Chem.MolFromSmarts('[#177]'), 4.317, 0),
    ('Condensed Aspartic Acid', Chem.MolFromSmarts('[#174]'), 3.887, 0),
    ('Condensed Arginine', Chem.MolFromSmarts('[#172]'), 12.503, 1),
    ('Condensed Lysine', Chem.MolFromSmarts('[#182]'), 10.517, 1),
    # ('Condensed AA end amino subst.', Chem.MolFromSmarts('[#0D1]'),9.15,1),
    # Can it somehow match up the wrong end? with a protection group??
    ('Condensed AA end amino', Chem.MolFromSmarts(
        '[#171,#172,#173,#174,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193;D1]'),
     9.564, 1),
    # ('Condensed AA end Carboxylic Subst.', Chem.MolFromSmarts('[$([#0][OH])]'),2.15,0), #All acidic/basic should be replaced with glycine so they can be removed??
    ('Condensed cys end amino', Chem.MolFromSmarts('[#175;D2]'), 9.564, 1),
    # ('Condensed AA end Carboxylic Subst.', Chem.MolFromSmarts('[$([#0][OH])]'),2.15,0), #All acidic/basic should be replaced with glycine so they can be removed??
    ('Condensed AA end Carboxylic', Chem.MolFromSmarts(
        '[$([#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193][OH])]'),
     2.383, 0)
]

# EMBOSS iep. 2017, http://emboss.sourceforge.net/apps/release/6.6/emboss/apps/iep.html.
pkatable_EMBOSS = [
    ('Tetramethylrhodamine #3',
     Chem.MolFromSmiles('CN(C)C1=C-C=C2-C(C4=C(C(=O)O)C=CC=C4)=C3-C=C-C(=[N+](C)C)C=C3OC2=C1'), [-10, 1.19, 3.59, 10],
     [1, 1, 0, 1]),
    # Guess DOI: 10.1039/C0OB01045F
    ('6-hydroxy-3H-xanthen-3-one', Chem.MolFromSmarts('c2c3ccc(O)cc3oc3cc(=O)ccc2-3'), [2.94, 6.56], [1, 0]),
    ('Anilin', Chem.MolFromSmarts('[$([NH2]c1ccccc1)]'), 4.87, 1),
    ('N-acetyl Hydrazine', Chem.MolFromSmarts('[$([NH2]NC(=O))]'), 2.81, 1),  # Reaxys
    ('Hydrazine', Chem.MolFromSmarts('[$([NH2]N)]'), 8.1, 1),  # Wikiped
    ('N-acetyl Piperazine', Chem.MolFromSmarts('[N+&!H0,NX3&H0;$(N1(C)CCN(C(=O)C)CC1)]'), 7.1, 1),  # Reaxys
    ('Piperazine', Chem.MolFromSmarts('C1CNCCN1'), 9.73, 1),  # Wikiped
    ('Pyridine', Chem.MolFromSmarts('n1ccccc1'), 5.25, 1),  # Wikiped
    ('Sulfonate', Chem.MolFromSmarts('[$([OH]S(=O)=O)]'), -1.5, 0),  # Guess
    ('Phosphonate', Chem.MolFromSmarts('[OH]P([OH])(=O)[#6]'), [1.5, 7.0], [0, -1]),  # Guess
    ('7H-Pyrimido[4,5-b][1,4]oxazine-2,4-diamine', Chem.MolFromSmarts('C1C=NC2=CN=C(N=C2O1)N'), 3.6, 1),  # PKA Guessed
    ('Tyrosine', Chem.MolFromSmarts('[$([O-,OH]c1ccccc1)]'), 10.1, 0),
    ('Histidine', Chem.MolFromSmarts(
        '[CH2X4][#6X3]1:[$([#7X3H+,#7X2H0+0]:[#6X3H]:[#7X3H]),$([#7X3H])]:[#6X3H]:[$([#7X3H+,#7X2H0+0]:[#6X3H]:[#7X3H]),$([#7X3H])]:[#6X3H]1'),
     6.5, 1),
    ('Histidine1', Chem.MolFromSmarts('[$(cC)]1cnc[nH]1'), 6.5, 1),
    ('Histidine2', Chem.MolFromSmarts('[$(cC)]1c[nH]cn1'), 6.5, 1),
    ('Glutamic Acid', Chem.MolFromSmarts('[$([CX3][CD2][CD2]C(N)C(=O))](=O)[OX1H0-,OX2H1]'), 4.1, 0),
    ('Cysteine', Chem.MolFromSmarts('[$([S-,SH][CD2]C(N)C(=O))]'), 8.5, 0),
    ('Condensed Cysteine', Chem.MolFromSmarts('[$([SH][#175])]'), 8.5, 0),  # Sidechain treated as normal atom
    ('Condensed Cysteine', Chem.MolFromSmarts('[$([SH][#178])]'), 8.5, 0),  # Sidechain treated as normal atom
    # ('Aspartic Acid', Chem.MolFromSmarts('[$([CX3][CD2])](=O)[OX1H0-,OX2H1]'),3.90,0),
    ('Aspartic Acid v2', Chem.MolFromSmarts('[$([CX3][CD2]C(N)C(=O))](=O)[OX1H0-,OX2H1]'), 3.9, 0),
    ('Arginine', Chem.MolFromSmarts('[$(N[CD2][CD2])]C(=N)N'), 12.5, 1),
    ('Lysine', Chem.MolFromSmarts('[$([ND1][CD2][CD2][CD2][CD2]C(N)C(=O))]'), 10.8, 1),
    ('End Carboxylate (avg)', Chem.MolFromSmarts('[$([CX3]CN)](=O)[OX1H0-,OX2H1]'), 3.6, 0),
    # ('End Amine (avg)',Chem.MolFromSmarts('[$([NX3;H2,H1;!$(NC=O)]CC=O)]'),9.15,1),
    ('End Amine (avg) PseudoAtom safe', Chem.MolFromSmarts(
        '[$([NX3;H2,H1;!$(NC=O);!$(N[#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193])]CC=O)]'),
     8.6, 1),
    ('Acrydine', Chem.MolFromSmarts('c1c2ccccc2nc2ccccc21'), 6.15, 1),
    ('Generic Carboxylic Acid', Chem.MolFromSmarts('[CX3](=O)[OX1H0-,OX2H1]'), 4.75, 0),
    (
    'Generic tert. Amine, not amide', Chem.MolFromSmarts('[N+&!H0&NX3,NX3;$(N(C)(C)C);!$(NC=[!#6]);!$(NC#[!#6])]'), 9.8,
    1),
    ('Generic Amine, not amide', Chem.MolFromSmarts(
        '[N+,NX3;!H0;!$(NC=[!#6]);!$(NC#[!#6])!$(N[#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193])]'),
     10., 1),
]

condensed_EMBOSS = [
    ('Condensed Tyrosine', Chem.MolFromSmarts('[#189]'), 10.1, 0),
    ('Condensed Histidine', Chem.MolFromSmarts('[#179]'), 6.5, 1),
    ('Condensed Glutamic Acid', Chem.MolFromSmarts('[#177]'), 4.4, 0),
    ('Condensed Aspartic Acid', Chem.MolFromSmarts('[#174]'), 3.9, 0),
    ('Condensed Arginine', Chem.MolFromSmarts('[#172]'), 12.5, 1),
    ('Condensed Lysine', Chem.MolFromSmarts('[#182]'), 10.8, 1),
    # ('Condensed AA end amino subst.', Chem.MolFromSmarts('[#0D1]'),9.15,1),
    # Can it somehow match up the wrong end? with a protection group??
    ('Condensed AA end amino', Chem.MolFromSmarts(
        '[#171,#172,#173,#174,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193;D1]'),
     8.6, 1),
    # ('Condensed AA end Carboxylic Subst.', Chem.MolFromSmarts('[$([#0][OH])]'),2.15,0), #All acidic/basic should be replaced with glycine so they can be removed??
    ('Condensed cys end amino', Chem.MolFromSmarts('[#175;D2]'), 8.6, 1),
    # ('Condensed AA end Carboxylic Subst.', Chem.MolFromSmarts('[$([#0][OH])]'),2.15,0), #All acidic/basic should be replaced with glycine so they can be removed??
    ('Condensed AA end Carboxylic', Chem.MolFromSmarts(
        '[$([#171,#172,#173,#174,#175,#176,#177,#178,#179,#180,#181,#182,#183,#184,#185,#186,#187,#188,#189,#190,#191,#192,#193][OH])]'),
     3.6, 0)
]

# Defaut pKa_tables
pkatable = pkatable_grimsley
# pkatable = pkatable_ipc
# pkatable = pkatable_EMBOSS

condensedtable = condensed_grimsley

def find_pKas(mol, pkatable=pkatable, condensedtable=condensedtable, debug=False, returnindices=False):
    """This function finds pKa values in a supplied molecule by matching it up to SMARTS rules in a table
    Returns the list of pKa values and a list with the charge of the group at pH levels blow the pKa.
    Example: So histidine has a +1 charge at pH levels below the pKa of ~6"""
    chargelist = []
    pkalist = []
    dummy = Chem.MolFromSmarts('[#178]')  # Used Gly instead of #0 or *, for compatibility with building blocks

    if returnindices:
        atomindices = []
        for atom in mol.GetAtoms():
            atom.SetProp('O_i', str(atom.GetIdx()))  # Save the Original Idx as a property

    # Look for condensed AA atoms with potential pKa values. Substitute with Glycine
    for s in condensedtable:
        if mol.HasSubstructMatch(s[1]):
            if debug: print
            "Match: %s" % (s[0]),
            # How many??
            substructmatches = mol.GetSubstructMatches(s[1])
            n = len(substructmatches)
            if debug: print
            "times %s" % (n)
            pkalist += [s[2]] * n
            chargelist += [s[3]] * n
            if returnindices:
                for match in substructmatches:
                    atomindices += [[int(mol.GetAtomWithIdx(idx).GetProp('O_i')) for idx in
                                     match]]  # Retrieve the original index of matches
            mol = AllChem.ReplaceSubstructs(mol, s[1], dummy)[
                0]  # Delete acidic/basic groups to prevent multiple matching

    # Look for molecular fragments with known pKa's
    for s in pkatable:
        # print s[0]
        if mol.HasSubstructMatch(s[1]):
            if debug: print
            "Match: %s" % (s[0]),
            # How many??
            substructmatches = mol.GetSubstructMatches(s[1])
            n = len(substructmatches)
            if debug: print
            "times %s" % (n)
            if type(s[2]) == type(float()):
                pkalist += [s[2]] * n
                chargelist += [s[3]] * n
            else:  # If there's multiple pKa values found for the molecular substructure
                for i in range(len(s[2])):
                    pkalist += [s[2][i]] * n
                    chargelist += [s[3][i]] * n
            if returnindices:
                for match in substructmatches:
                    atomindices += [
                        [int(mol.GetAtomWithIdx(idx).GetProp('O_i')) for idx in match]]  # Retrieve the original i
            mol = AllChem.ReplaceSubstructs(mol, s[1], dummy)[0]
    if returnindices:
        return pkalist, chargelist, atomindices
    return pkalist, chargelist


def charge(ph, pkalist, chargelist):
    """Jacob Tolborgs charge model where the charge is assigned from partial charges from all pKa values at the pH point"""
    chargesum = []
    abschargesum = []
    for charge, pka in zip(chargelist, pkalist):
        # print charge, pka
        if charge == 1:
            charge = 1 / (1 + 10 ** (ph - pka))
            chargesum.append(charge)
        else:
            charge = -1 / (1 + 10 ** -(ph - pka))
            chargesum.append(charge)
            abschargesum.append(abs(charge))
    return sum(chargesum), sum(abschargesum)


def pI(pkalist, chargelist):
    """Uses Jacob Tolborgs charge function and calculates pI.
    If only only acidic or basic groups are found, the pI is ill defined and returned as 42 or -42"""
    # Check if Only acidic or basic groups are present.
    if len(unique(chargelist)) == 0:  # No pKa groups present.
        pI = 7
    elif len(unique(chargelist)) == 1:  # Only one type is present
        if chargelist[0] == 0:  # Only acidic groups are present
            pI = 3
        elif chargelist[0] == 1:  # Only basic groups are present
            pI = 11
    else:
        # Find pI by simulation in the pH range 0 to 0
        chargecol = []
        for i in range(0, 1400):
            ph = i / 100.
            chargecol.append(charge(ph, pkalist, chargelist))  # Calculate charge
        pI = argmin(
            np_abs(chargecol)) / 100.  # Simply taking the smallest absolute value, and dividing the index with 100
    # print "pI %.1f"%pI
    return pI

from olorenchemengine.representations import BaseCompoundVecRepresentation
import numpy as np
from olorenchemengine.base_class import log_arguments
class calc_pI(BaseCompoundVecRepresentation):
    @log_arguments
    def __init__(self, ph=7.4, log=True, **kwargs):
        self.ph=ph
        super().__init__(names=["pI", f"Charge{self.ph}", f"TotalAbsCharge{self.ph}"], log=False, **kwargs)

    def _convert(self, smiles, y=None):
        # find pKa values and charge class
        m = Chem.MolFromSmiles(smiles)
        pkalist, chargelist = find_pKas(m)
        # Calculate pI
        pIpred = pI(pkalist, chargelist)
        return np.array([pIpred, *charge(self.ph, pkalist, chargelist)])


""" Wraps the model presented in `Strategies for pre-training graph neural networks <https://arxiv.org/abs/1905.12265>`_

`GitHub repository <https://github.com/snap-stanford/pretrain-gnns>`_
"""

from olorenchemengine.internal import mock_imports

try:
    from torch_geometric.data import DataLoader
except ImportError:
    mock_imports(globals(), "DataLoader")

try:
    import torch
    import torch.nn as nn
    import torch.optim as optim
except ImportError:
    mock_imports(globals(), "torch", "nn", "optim")


from tqdm import tqdm
import numpy as np

from .model import GNN_graphpred, GNN, global_add_pool, global_mean_pool, global_max_pool, GlobalAttention, Set2Set
from sklearn.metrics import roc_auc_score

import olorenchemengine as oce
from olorenchemengine.representations import AtomFeaturizer, BaseRepresentation, BondFeaturizer, TorchGeometricGraph, BaseVecRepresentation

import os
import io
from typing import Union, List, Tuple, Dict, Optional

from olorenchemengine.base_class import BaseModel, log_arguments, QuantileTransformer
from olorenchemengine.internal import download_public_file

from rdkit import Chem

allowable_features = {
    'possible_atomic_num_list' : list(range(1, 119)),
    'possible_formal_charge_list' : [-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5],
    'possible_chirality_list' : [
        Chem.rdchem.ChiralType.CHI_UNSPECIFIED,
        Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
        Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW,
        Chem.rdchem.ChiralType.CHI_OTHER
    ],
    'possible_hybridization_list' : [
        Chem.rdchem.HybridizationType.S,
        Chem.rdchem.HybridizationType.SP, Chem.rdchem.HybridizationType.SP2,
        Chem.rdchem.HybridizationType.SP3, Chem.rdchem.HybridizationType.SP3D,
        Chem.rdchem.HybridizationType.SP3D2, Chem.rdchem.HybridizationType.UNSPECIFIED
    ],
    'possible_numH_list' : [0, 1, 2, 3, 4, 5, 6, 7, 8],
    'possible_implicit_valence_list' : [0, 1, 2, 3, 4, 5, 6],
    'possible_degree_list' : [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    'possible_bonds' : [
        Chem.rdchem.BondType.SINGLE,
        Chem.rdchem.BondType.DOUBLE,
        Chem.rdchem.BondType.TRIPLE,
        Chem.rdchem.BondType.AROMATIC
    ],
    'possible_bond_dirs' : [ # only for double bond stereo information
        Chem.rdchem.BondDir.NONE,
        Chem.rdchem.BondDir.ENDUPRIGHT,
        Chem.rdchem.BondDir.ENDDOWNRIGHT
    ]
}

class SPGNN_AF(AtomFeaturizer):
    @property
    def length(self):
        return 2

    def convert(self, atom: Chem.Atom):
        atom_feature = [allowable_features['possible_atomic_num_list'].index(
            atom.GetAtomicNum())] + [allowable_features[
            'possible_chirality_list'].index(atom.GetChiralTag())]
        x = np.array(atom_feature)
        return x

class SPGNN_BF(BondFeaturizer):
    @property
    def length(self):
        return 2

    def convert(self, bond: Chem.Bond):
        edge_feature = [allowable_features['possible_bonds'].index(
                bond.GetBondType())] + [allowable_features[
                                            'possible_bond_dirs'].index(
                bond.GetBondDir())]
        x = np.array(edge_feature)
        return x

class SPGNN_PYG(TorchGeometricGraph):
    @log_arguments
    def __init__(self):
        super().__init__(SPGNN_AF(), SPGNN_BF(), log = False)

class SPGNNVecRep(BaseVecRepresentation):
    """SPGNN_REP gives the output of the model presented in `Strategies for pre-training graph neural networks <https://arxiv.org/abs/1905.12265>`_
    `GitHub repository <https://github.com/snap-stanford/pretrain-gnns>` as a molecular representation.

    Attributes:
        available_pretrained_models (List[str]): List of available pretrained models; passed in the model_type parameter

    Parameters:
        model_type (str): Type of model to use; default: "contextpred"
    """

    available_pretrained_models = ["contextpred",
        "edgepred",
        "infomax",
        "masking",
        "supervised_contextpred",
        "supervised_edgepred",
        "supervised_infomax",
        "supervised_masking",
        "supervised",
        "gat_supervised_contextpred",
        "gat_supervised",
        "gat_contextpred"]

    @log_arguments
    def __init__(self, model_type = "contextpred",
        batch_size = 32,
        epochs = 100,
        lr = 0.001,
        lr_scale = 1,
        decay = 0,
        num_layer = 5,
        emb_dim = 300,
        dropout_ratio = 0.5,
        graph_pooling = "mean",
        JK = "last",
        gnn_type = "gin", **kwargs):

        self.model_type = model_type

        if "gat" in model_type:
            gnn_type = "gat"
            emb_dim= 300

        self.representation = SPGNN_PYG()

        self.batch_size = batch_size
        self.epochs = epochs
        self.num_workers = oce.CONFIG["NUM_WORKERS"]
        self.map_location = oce.CONFIG["MAP_LOCATION"]
        self.device = oce.CONFIG["DEVICE"]

        self.model = GNN(num_layer, emb_dim, JK, dropout_ratio, gnn_type = gnn_type)

        input_model_file = download_public_file(f"SPGNN_saves/{self.model_type}.pth")

        self.model.load_state_dict(torch.load(input_model_file, map_location=self.map_location))
        self.model.eval()
        self.model.to(self.device)

        if graph_pooling == "sum":
            self.pool = global_add_pool
        elif graph_pooling == "mean":
            self.pool = global_mean_pool
        elif graph_pooling == "max":
            self.pool = global_max_pool
        elif graph_pooling == "attention":
            if self.JK == "concat":
                self.pool = GlobalAttention(gate_nn = torch.nn.Linear((self.num_layer + 1) * emb_dim, 1))
            else:
                self.pool = GlobalAttention(gate_nn = torch.nn.Linear(emb_dim, 1))
        elif graph_pooling[:-1] == "set2set":
            set2set_iter = int(graph_pooling[-1])
            if self.JK == "concat":
                self.pool = Set2Set((self.num_layer + 1) * emb_dim, set2set_iter)
            else:
                self.pool = Set2Set(emb_dim, set2set_iter)
        else:
            raise ValueError("Invalid graph pooling type.")

    def _convert(self, smiles: str, y: Union[int, float, np.number] = None) -> np.ndarray:
        assert Exception, "Please directly use convert method"

    def convert(self, smiles, **kwargs):
        X = self.representation.convert(smiles)
        loader = DataLoader(X, batch_size=self.batch_size,
            shuffle=False, num_workers=self.num_workers)

        self.model.eval()
        y_pred = []

        for step, batch in enumerate(tqdm(loader, desc="Iteration")):
            batch = batch.to(self.device)

            with torch.no_grad():
                pred = self.model(batch.x, batch.edge_index, batch.edge_attr)
                pred = self.pool(pred, batch.batch).detach().cpu().numpy()

            y_pred.append(pred)
        return np.concatenate(y_pred)

def train(model, device, loader, optimizer, setting):
    if setting == "classification":
        criterion = nn.BCEWithLogitsLoss(reduction = "none")
    else:
        criterion = nn.MSELoss()
    model.train()

    for step, batch in enumerate(loader):
        batch = batch.to(device)
        pred = model(batch.x, batch.edge_index, batch.edge_attr, batch.batch)
        y = batch.y.view(pred.shape).to(torch.float64)

        #Loss matrix
        loss_mat = criterion(pred.double(), y)

        optimizer.zero_grad()

        loss = torch.mean(loss_mat)
        loss.backward()

        optimizer.step()

def predict(model, device, loader, setting = "classification"):
    model.eval()
    y_pred = []

    for step, batch in enumerate(tqdm(loader, desc="Iteration")):
        batch = batch.to(device)

        with torch.no_grad():
            pred = model(batch.x, batch.edge_index, batch.edge_attr, batch.batch)

        y_pred.append(pred)
    if setting == "classification":
        y_pred = nn.Sigmoid()(torch.cat(y_pred, dim = 0)).cpu().numpy()
    else:
        y_pred = torch.cat(y_pred, dim = 0).cpu().numpy()
    return y_pred.flatten()

class SPGNN(BaseModel):
    """SPGNN is the model presented in `Strategies for pre-training graph neural networks <https://arxiv.org/abs/1905.12265>`_
    `GitHub repository <https://github.com/snap-stanford/pretrain-gnns>`_

    Attributes:
        available_pretrained_models (List[str]): List of available pretrained models; passed in the model_type parameter

    Parameters:
        model_type (str): Type of model to use; default: "contextpred"
    """

    available_pretrained_models = ["contextpred",
        "edgepred",
        "infomax",
        "masking",
        "supervised_contextpred",
        "supervised_edgepred",
        "supervised_infomax",
        "supervised_masking",
        "supervised",
        "gat_supervised_contextpred",
        "gat_supervised",
        "gat_contextpred"]

    @log_arguments
    def __init__(self, model_type = "contextpred",
        batch_size = 32,
        epochs = 100,
        lr = 0.001,
        lr_scale = 1,
        decay = 0,
        num_layer = 5,
        emb_dim = 300,
        dropout_ratio = 0.5,
        graph_pooling = "mean",
        JK = "last",
        gnn_type = "gin", **kwargs):

        self.model_type = model_type

        if "gat" in model_type:
            gnn_type = "gat"
            emb_dim= 300
        self.representation = SPGNN_PYG()

        self.batch_size = batch_size
        self.epochs = epochs

        self.num_workers = oce.CONFIG["NUM_WORKERS"]
        self.map_location = oce.CONFIG["MAP_LOCATION"]
        self.device = oce.CONFIG["DEVICE"]

        self.model = GNN_graphpred(num_layer,
            emb_dim,
            1,
            JK = JK,
            drop_ratio = dropout_ratio,
            graph_pooling = graph_pooling,
            gnn_type = gnn_type)

        input_model_file = download_public_file(f"SPGNN_saves/{self.model_type}.pth")

        self.model.from_pretrained(input_model_file, map_location=self.map_location)

        super().__init__(**kwargs)

        self.model.to(self.device)

        model_param_group = []
        model_param_group.append({"params": self.model.gnn.parameters()})
        if graph_pooling == "attention":
            model_param_group.append({"params": self.model.pool.parameters(), "lr":lr*lr_scale})
        model_param_group.append({"params": self.model.graph_pred_linear.parameters(), "lr":lr* lr_scale})

        self.optimizer = optim.Adam(model_param_group, lr=lr, weight_decay=decay)
        self.sigmoid = nn.Sigmoid()

    def preprocess(self, X, y, **kwargs):
        if y is None:
            y = [None]*len(X)
        return self.representation.convert(X, ys=y)

    def _fit(self, X, y, **kwargs):
        dataloader = DataLoader(X, batch_size=self.batch_size, shuffle=True, num_workers=self.num_workers)
        for epoch in range(self.epochs):
            train(self.model, self.device, dataloader, self.optimizer, self.setting)

    def _predict(self, X, **kwargs):
        dataloader = DataLoader(X, batch_size=self.batch_size, shuffle=False, num_workers=self.num_workers)
        return predict(self.model, self.device, dataloader, setting = self.setting)

    def _save(self) -> str:
        d = super()._save()
        buffer = io.BytesIO()
        torch.save(self.model, buffer)
        d.update({"save": buffer.getvalue()})
        return d

    def _load(self, d):
        super()._load(d)
        self.model = torch.load(io.BytesIO(d["save"]))

    @classmethod
    def AllInstances(cls):
        return [cls(model_type=mt) for mt in cls.available_pretrained_models]


"""
Model comparisons is a set of visualizations that compare various models or
parameters thereof. This can be used for hyperparameter tuning, benchmarking
visualizations, or model selections.
"""

from olorenchemengine.base_class import *
from olorenchemengine.representations import *
from olorenchemengine.dataset import *
from olorenchemengine.uncertainty import *
from olorenchemengine.interpret import *
from olorenchemengine.internal import *
from olorenchemengine.manager import *
from olorenchemengine.visualizations import *

class VisualizationModelManager_Bar(BaseVisualization):
    """ Visualize the database of a model manager in a bar plot, with the model
    parameters and name displayed on hover

    Parameters:
        mm (ModelManager): The model manager to visualize
        top (int): The number of models to display
    """

    @log_arguments
    def __init__(self, mm: BaseModelManager, top = 5, log=True, **kwargs):
        self.df = mm.get_model_database()
        self.metrics = mm.metrics
        self.mm = mm
        self.top = top
        super().__init__(log=False)
        self.packages += ["plotly"]

    @property
    def JS_NAME(self) -> str:
        return "ModelManagerBar"

    def get_data(self, metric = None) -> dict:
        """
        Get the data for the visualization

        Parameters:
            metric (str): The metrics to use for the visualization

        Returns:
            dict: The data for the visualization
        """
        df2 = self.df.sort_values(by=metric, ascending=(True if self.mm.metric_direction[metric] == "lower" else False))
        df2 = df2.iloc[:self.top]
        d =  df2.to_dict("l")
        if not metric is None:
            assert metric in self.metrics, f"Designated metric, {metric}, not in model manager metrics: {self.metrics}"
            d["metric"] = metric
        else:
            d["metric"] = self.metrics[0]
        d["title"] = "Bar Plot of Best Models"
        return d

    def render_ipynb(self, metric=None, *args, print_html_js=False, **kwargs) -> str:
        """
        Render the visualization in an ipython notebook

        Parameters:
            metric (str): The metric to use for the visualization
            *args: Additional arguments to pass to the visualization
            print_html_js (bool): Whether to print the javascript code for the visualization
            **kwargs: Additional arguments to pass to the visualization

        Returns:
            str: The html code for the visualization
        """
        d = self.get_data(metric)
        return self.render_ipynb_from_dict(d, *args, print_html_js=print_html_js, **kwargs)

class ModelOverTime(CompoundScatterPlot):
    """ Visualizes the performance of a given model architecture over a length of time.

    This visualization will use a date split and for each date i, will train a model
    on the data on dates 1, ..., i-1, evaluating on date i. Note, any existing splits
    in the dataset will be ignored

    Parameters:
        dataset (BaseDataset): The dataset to use for the visualization
        model (BaseModel): The model to use for the visualization
    """

    @log_arguments
    def __init__(self, dataset: BaseDataset, model: BaseModel, *args, log=True,
        title="Model Performance over Time", yaxis_title="Error", xaxis_title="Date", **kwargs):
        self.dataset = dataset
        self.model = model
        self.title = title
        self.yaxis_title = yaxis_title
        self.xaxis_title = xaxis_title

        self.dataset.data = self.dataset.data.sort_values(by=[self.dataset.date_col])
        self.dates = self.dataset.data[self.dataset.date_col].unique()
        results = pd.DataFrame()
        for i in tqdm(range(len(self.dates)-1)):
            # try:
            train_date = self.dates[i]
            test_date = self.dates[i+1]
            train = self.dataset.data[self.dataset.data[self.dataset.date_col] <= train_date]
            train["split"] = "train"
            test = self.dataset.data[self.dataset.data[self.dataset.date_col] == test_date]
            test["split"] = "test"
            DATA = pd.concat([train, test])
            dataset = (oce.BaseDataset(data = DATA.to_csv(), structure_col="SMILES", property_col = "log AUC") + oce.CleanStructures())
            print(f"Train size is {len(train)} and test size is {len(test)}")
            self.model.fit(*dataset.train_dataset)

            test_set = dataset.data[dataset.data["split"] == "test"]
            test_set["pred"] = self.model.predict(test_set["SMILES"])

            results = pd.concat([results, test_set])

            if self.model.setting == "regression":
                results["Error (|Predicted - True|)"] = np.abs(results["pred"] - results[self.dataset.property_col])
                self.yaxis_title = "Error (|Predicted - True|)"
            elif self.model.setting == "classification":
                results["Error (p or 1-p)"] = np.abs(results[self.dataset.property_col]+(-2*results[self.dataset.property_col]+1)*results["pred"])
                self.yaxis_title = "Error (p or 1-p)"

        results["X"] = results[self.dataset.date_col]
        if self.model.setting == "regression":
            results["Y"] = results["Error (|Predicted - True|)"]
            self.xaxis_title = xaxis_title
        elif self.model.setting == "classification":
            results["Y"] = results["Error (p or 1-p)"]
            self.yaxis_title = yaxis_title
        results["SMILES"] = results[self.dataset.structure_col]
        super().__init__(*args,df = results, xaxis_type="date",
            title=title, xaxis_title=self.xaxis_title, yaxis_title=self.yaxis_title, log=False, **kwargs)

    def get_data(self):
        d = super().get_data()
        window = 10
        quantile = 0.80
        results = list(zip(pd.to_datetime(self.df["X"]).apply(lambda x: x.timestamp()), self.df["Y"]))
        scores, residuals = zip(*sorted(results))
        X = pd.Series(scores).rolling(window).mean()
        y = pd.Series(residuals).rolling(window).quantile(quantile)
        X = np.array(X[~np.isnan(X)])
        y = np.array(y[~np.isnan(y)])

        d["trace_update"] = {'x':self.df["X"].tolist(),
            'y': y.tolist(),
            'mode': 'markers',
            'type': 'scatter',
            'marker': {'color': 'red'}}
        return d